import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from typing import List, Tuple
from scipy.stats import norm
import os


from vega_sim.reinforcement.la_market_state import LAMarketState, states_to_sarsa
from vega_sim.reinforcement.agents.learning_agent_MO_with_vol import Action


def action_to_vector(action: Action):
    if action.sell:
        output = 0
    elif action.buy:
        output = 1
    else:
        output = 2
    return output


def plot_simulation(
    simulation: List[Tuple[LAMarketState, Action]], results_dir: str, tag: int
):
    """Plot the states of a learning_agent in the simulation"""
    sars = states_to_sarsa(simulation)
    reward = np.array([_sars[2] for _sars in sars])
    next_price_list = [_sars[0].next_price for _sars in sars]
    next_price_list.insert(0, next_price_list[0])
    next_price = np.array(next_price_list)
    best_bid = np.array([_sars[0].bid_prices[0] for _sars in sars])
    best_bid[-1] = best_bid[-2]  # because at settlement it's `0` and messes up the plot
    best_ask = np.array([_sars[0].ask_prices[0] for _sars in sars])
    best_ask[-1] = best_ask[-2]  # because at settlement it's `0` and messes up the plot
    position = np.array([_sars[0].position for _sars in sars])

    total_balance = np.array([_sars[0].full_balance for _sars in sars])
    # action_discr = np.array([action_to_vector(_sars[1]) for _sars in sars])
    # action_volume = np.array([_sars[1].volume for _sars in sars])
    # t = np.linspace(1, len(action_volume), len(action_volume))

    fig, ax = plt.subplots(nrows=2, ncols=2, figsize=(12, 8))
    fig = plt.figure(figsize=(12, 12))
    spec = fig.add_gridspec(3, 2)

    ax0 = fig.add_subplot(spec[0, 0])
    ax0.plot(best_bid, label="best bid")
    ax0.plot(best_ask, label="best ask")
    ax0.plot(next_price, label="next_price")
    ax0.legend()

    ax1 = fig.add_subplot(spec[0, 1])
    ax1.plot(position, "x", label="position")
    ax1.legend()

    ax2 = fig.add_subplot(spec[1, 0])
    ax2.plot(total_balance, ".", label="total balance")
    ax2.set_yscale("log")
    ax2.legend()

    # ax3 = fig.add_subplot(spec[1, 1])
    # scatter = ax3.scatter(t, action_volume, c=action_discr)
    # produce a legend with the unique colors from the scatter
    # handles, labels = scatter.legend_elements()
    # labels = ["buy", "sell", "do nothing"]
    # ax3.legend(*(handles, labels), title="Type of action")
    # ax3.set_ylabel("volume")

    ax4 = fig.add_subplot(spec[2, :])
    ax4.plot(reward.cumsum(), label="cumulative reward")
    ax4.legend()
    fig.savefig(os.path.join(results_dir, "sim{}.pdf".format(tag)))
    plt.close()


def plot_learning(results_dir: str, logfile_pol_imp: str, logfile_pol_eval: str):

    data = pd.read_csv(logfile_pol_imp)
    plt.figure()
    plt.plot(data["iteration"], data["loss"])
    plt.ylabel("KL(policy|e^Q)")
    plt.xlabel("Iteration")
    plt.title("Policy improvement")

    plt.savefig(os.path.join(results_dir, "learn_pol_imp.pdf"))
    plt.close()

    data = pd.read_csv(logfile_pol_eval)
    plt.figure()
    plt.plot(data["iteration"], np.log(data["loss"]))
    plt.ylabel("log: Bellman error est")
    plt.xlabel("Iteration")
    plt.title("Q-function estimation")
    plt.savefig(os.path.join(results_dir, "learn_pol_eval.pdf"))
    plt.close()


def plot_pnl(results_dir: str, logfile_pnl: str):
    data = pd.read_csv(logfile_pnl)
    pnl = data["pnl"].to_numpy()
    plt.figure()
    n, bins, patches = plt.hist(x=pnl, bins=20, color="#0504aa", alpha=0.7, rwidth=0.85)
    plt.grid(axis="y", alpha=0.75)
    plt.xlabel("PnL")
    plt.ylabel("Frequency")
    conf = 0.975
    err_bar = norm.ppf(conf) * pnl.std() / len(pnl)
    text = str(pnl.mean() - err_bar) + " < PnL < " + str(pnl.mean() + err_bar)
    plt.title(text)
    maxfreq = n.max()
    # Set a clean upper y-axis limit.
    plt.ylim(ymax=np.ceil(maxfreq / 10) * 10 if maxfreq % 10 else maxfreq + 10)
    plt.savefig(os.path.join(results_dir, "learn_pnl.pdf"))
    plt.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/governance.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import markets_pb2 as vega_dot_markets__pb2
from . import vega_pb2 as vega_dot_vega__pb2
from . import assets_pb2 as vega_dot_assets__pb2
from . import data_source_pb2 as vega_dot_data__source__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15vega/governance.proto\x12\x04vega\x1a\x12vega/markets.proto\x1a\x0fvega/vega.proto\x1a\x11vega/assets.proto\x1a\x16vega/data_source.proto"\xcf\x03\n\rFutureProduct\x12)\n\x10settlement_asset\x18\x01 \x01(\tR\x0fsettlementAsset\x12\x1d\n\nquote_name\x18\x02 \x01(\tR\tquoteName\x12i\n$data_source_spec_for_settlement_data\x18\x03 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR\x1f\x64\x61taSourceSpecForSettlementData\x12q\n(data_source_spec_for_trading_termination\x18\x04 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR#dataSourceSpecForTradingTermination\x12\\\n\x18\x64\x61ta_source_spec_binding\x18\x05 \x01(\x0b\x32#.vega.DataSourceSpecToFutureBindingR\x15\x64\x61taSourceSpecBinding\x12\x38\n\x18settlement_data_decimals\x18\x06 \x01(\rR\x16settlementDataDecimals"{\n\x17InstrumentConfiguration\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x12\n\x04\x63ode\x18\x02 \x01(\tR\x04\x63ode\x12-\n\x06\x66uture\x18\x64 \x01(\x0b\x32\x13.vega.FutureProductH\x00R\x06\x66utureB\t\n\x07product"\xa1\x04\n\x16NewMarketConfiguration\x12=\n\ninstrument\x18\x01 \x01(\x0b\x32\x1d.vega.InstrumentConfigurationR\ninstrument\x12%\n\x0e\x64\x65\x63imal_places\x18\x02 \x01(\x04R\rdecimalPlaces\x12\x1a\n\x08metadata\x18\x03 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x04 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12k\n\x1fliquidity_monitoring_parameters\x18\x05 \x01(\x0b\x32#.vega.LiquidityMonitoringParametersR\x1dliquidityMonitoringParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormal\x12\x36\n\x17position_decimal_places\x18\x06 \x01(\x03R\x15positionDecimalPlacesB\x11\n\x0frisk_parameters"C\n\tNewMarket\x12\x36\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1c.vega.NewMarketConfigurationR\x07\x63hanges"f\n\x0cUpdateMarket\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12\x39\n\x07\x63hanges\x18\x02 \x01(\x0b\x32\x1f.vega.UpdateMarketConfigurationR\x07\x63hanges"\xcb\x03\n\x19UpdateMarketConfiguration\x12\x43\n\ninstrument\x18\x01 \x01(\x0b\x32#.vega.UpdateInstrumentConfigurationR\ninstrument\x12\x1a\n\x08metadata\x18\x02 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x03 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12k\n\x1fliquidity_monitoring_parameters\x18\x04 \x01(\x0b\x32#.vega.LiquidityMonitoringParametersR\x1dliquidityMonitoringParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormalB\x11\n\x0frisk_parameters"s\n\x1dUpdateInstrumentConfiguration\x12\x12\n\x04\x63ode\x18\x01 \x01(\tR\x04\x63ode\x12\x33\n\x06\x66uture\x18\x64 \x01(\x0b\x32\x19.vega.UpdateFutureProductH\x00R\x06\x66utureB\t\n\x07product"\xaa\x03\n\x13UpdateFutureProduct\x12\x1d\n\nquote_name\x18\x01 \x01(\tR\tquoteName\x12i\n$data_source_spec_for_settlement_data\x18\x02 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR\x1f\x64\x61taSourceSpecForSettlementData\x12q\n(data_source_spec_for_trading_termination\x18\x03 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR#dataSourceSpecForTradingTermination\x12\\\n\x18\x64\x61ta_source_spec_binding\x18\x04 \x01(\x0b\x32#.vega.DataSourceSpecToFutureBindingR\x15\x64\x61taSourceSpecBinding\x12\x38\n\x18settlement_data_decimals\x18\x05 \x01(\rR\x16settlementDataDecimals"J\n\x16UpdateNetworkParameter\x12\x30\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x16.vega.NetworkParameterR\x07\x63hanges"8\n\x08NewAsset\x12,\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x12.vega.AssetDetailsR\x07\x63hanges"\\\n\x0bUpdateAsset\x12\x19\n\x08\x61sset_id\x18\x01 \x01(\tR\x07\x61ssetId\x12\x32\n\x07\x63hanges\x18\x02 \x01(\x0b\x32\x18.vega.AssetDetailsUpdateR\x07\x63hanges"\r\n\x0bNewFreeform"\x90\x04\n\rProposalTerms\x12+\n\x11\x63losing_timestamp\x18\x01 \x01(\x03R\x10\x63losingTimestamp\x12/\n\x13\x65nactment_timestamp\x18\x02 \x01(\x03R\x12\x65nactmentTimestamp\x12\x31\n\x14validation_timestamp\x18\x03 \x01(\x03R\x13validationTimestamp\x12\x39\n\rupdate_market\x18\x65 \x01(\x0b\x32\x12.vega.UpdateMarketH\x00R\x0cupdateMarket\x12\x30\n\nnew_market\x18\x66 \x01(\x0b\x32\x0f.vega.NewMarketH\x00R\tnewMarket\x12X\n\x18update_network_parameter\x18g \x01(\x0b\x32\x1c.vega.UpdateNetworkParameterH\x00R\x16updateNetworkParameter\x12-\n\tnew_asset\x18h \x01(\x0b\x32\x0e.vega.NewAssetH\x00R\x08newAsset\x12\x36\n\x0cnew_freeform\x18i \x01(\x0b\x32\x11.vega.NewFreeformH\x00R\x0bnewFreeform\x12\x36\n\x0cupdate_asset\x18j \x01(\x0b\x32\x11.vega.UpdateAssetH\x00R\x0bupdateAssetB\x08\n\x06\x63hange"W\n\x11ProposalRationale\x12 \n\x0b\x64\x65scription\x18\x01 \x01(\tR\x0b\x64\x65scription\x12\x14\n\x05title\x18\x04 \x01(\tR\x05titleJ\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04"\x86\x03\n\x0eGovernanceData\x12*\n\x08proposal\x18\x01 \x01(\x0b\x32\x0e.vega.ProposalR\x08proposal\x12\x1c\n\x03yes\x18\x02 \x03(\x0b\x32\n.vega.VoteR\x03yes\x12\x1a\n\x02no\x18\x03 \x03(\x0b\x32\n.vega.VoteR\x02no\x12?\n\tyes_party\x18\x04 \x03(\x0b\x32".vega.GovernanceData.YesPartyEntryR\x08yesParty\x12<\n\x08no_party\x18\x05 \x03(\x0b\x32!.vega.GovernanceData.NoPartyEntryR\x07noParty\x1aG\n\rYesPartyEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x05value\x18\x02 \x01(\x0b\x32\n.vega.VoteR\x05value:\x02\x38\x01\x1a\x46\n\x0cNoPartyEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x05value\x18\x02 \x01(\x0b\x32\n.vega.VoteR\x05value:\x02\x38\x01"\x9a\x07\n\x08Proposal\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1c\n\treference\x18\x02 \x01(\tR\treference\x12\x19\n\x08party_id\x18\x03 \x01(\tR\x07partyId\x12*\n\x05state\x18\x04 \x01(\x0e\x32\x14.vega.Proposal.StateR\x05state\x12\x1c\n\ttimestamp\x18\x05 \x01(\x03R\ttimestamp\x12)\n\x05terms\x18\x06 \x01(\x0b\x32\x13.vega.ProposalTermsR\x05terms\x12\x30\n\x06reason\x18\x07 \x01(\x0e\x32\x13.vega.ProposalErrorH\x00R\x06reason\x88\x01\x01\x12(\n\rerror_details\x18\x08 \x01(\tH\x01R\x0c\x65rrorDetails\x88\x01\x01\x12\x35\n\trationale\x18\t \x01(\x0b\x32\x17.vega.ProposalRationaleR\trationale\x12\x35\n\x16required_participation\x18\n \x01(\tR\x15requiredParticipation\x12+\n\x11required_majority\x18\x0b \x01(\tR\x10requiredMajority\x12^\n)required_liquidity_provider_participation\x18\x0c \x01(\tH\x02R&requiredLiquidityProviderParticipation\x88\x01\x01\x12T\n$required_liquidity_provider_majority\x18\r \x01(\tH\x03R!requiredLiquidityProviderMajority\x88\x01\x01"\xae\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x10\n\x0cSTATE_FAILED\x10\x01\x12\x0e\n\nSTATE_OPEN\x10\x02\x12\x10\n\x0cSTATE_PASSED\x10\x03\x12\x12\n\x0eSTATE_REJECTED\x10\x04\x12\x12\n\x0eSTATE_DECLINED\x10\x05\x12\x11\n\rSTATE_ENACTED\x10\x06\x12\x1f\n\x1bSTATE_WAITING_FOR_NODE_VOTE\x10\x07\x42\t\n\x07_reasonB\x10\n\x0e_error_detailsB,\n*_required_liquidity_provider_participationB\'\n%_required_liquidity_provider_majority"\x91\x03\n\x04Vote\x12\x19\n\x08party_id\x18\x01 \x01(\tR\x07partyId\x12&\n\x05value\x18\x02 \x01(\x0e\x32\x10.vega.Vote.ValueR\x05value\x12\x1f\n\x0bproposal_id\x18\x03 \x01(\tR\nproposalId\x12\x1c\n\ttimestamp\x18\x04 \x01(\x03R\ttimestamp\x12\x43\n\x1etotal_governance_token_balance\x18\x05 \x01(\tR\x1btotalGovernanceTokenBalance\x12\x41\n\x1dtotal_governance_token_weight\x18\x06 \x01(\tR\x1atotalGovernanceTokenWeight\x12\x42\n\x1etotal_equity_like_share_weight\x18\x07 \x01(\tR\x1atotalEquityLikeShareWeight";\n\x05Value\x12\x15\n\x11VALUE_UNSPECIFIED\x10\x00\x12\x0c\n\x08VALUE_NO\x10\x01\x12\r\n\tVALUE_YES\x10\x02*\xd1\x0c\n\rProposalError\x12\x1e\n\x1aPROPOSAL_ERROR_UNSPECIFIED\x10\x00\x12&\n"PROPOSAL_ERROR_CLOSE_TIME_TOO_SOON\x10\x01\x12&\n"PROPOSAL_ERROR_CLOSE_TIME_TOO_LATE\x10\x02\x12&\n"PROPOSAL_ERROR_ENACT_TIME_TOO_SOON\x10\x03\x12&\n"PROPOSAL_ERROR_ENACT_TIME_TOO_LATE\x10\x04\x12&\n"PROPOSAL_ERROR_INSUFFICIENT_TOKENS\x10\x05\x12.\n*PROPOSAL_ERROR_INVALID_INSTRUMENT_SECURITY\x10\x06\x12\x1d\n\x19PROPOSAL_ERROR_NO_PRODUCT\x10\x07\x12&\n"PROPOSAL_ERROR_UNSUPPORTED_PRODUCT\x10\x08\x12"\n\x1ePROPOSAL_ERROR_NO_TRADING_MODE\x10\x0b\x12+\n\'PROPOSAL_ERROR_UNSUPPORTED_TRADING_MODE\x10\x0c\x12)\n%PROPOSAL_ERROR_NODE_VALIDATION_FAILED\x10\r\x12.\n*PROPOSAL_ERROR_MISSING_BUILTIN_ASSET_FIELD\x10\x0e\x12\x31\n-PROPOSAL_ERROR_MISSING_ERC20_CONTRACT_ADDRESS\x10\x0f\x12 \n\x1cPROPOSAL_ERROR_INVALID_ASSET\x10\x10\x12*\n&PROPOSAL_ERROR_INCOMPATIBLE_TIMESTAMPS\x10\x11\x12%\n!PROPOSAL_ERROR_NO_RISK_PARAMETERS\x10\x12\x12\x30\n,PROPOSAL_ERROR_NETWORK_PARAMETER_INVALID_KEY\x10\x13\x12\x32\n.PROPOSAL_ERROR_NETWORK_PARAMETER_INVALID_VALUE\x10\x14\x12\x36\n2PROPOSAL_ERROR_NETWORK_PARAMETER_VALIDATION_FAILED\x10\x15\x12\x35\n1PROPOSAL_ERROR_OPENING_AUCTION_DURATION_TOO_SMALL\x10\x16\x12\x35\n1PROPOSAL_ERROR_OPENING_AUCTION_DURATION_TOO_LARGE\x10\x17\x12/\n+PROPOSAL_ERROR_COULD_NOT_INSTANTIATE_MARKET\x10\x19\x12)\n%PROPOSAL_ERROR_INVALID_FUTURE_PRODUCT\x10\x1a\x12)\n%PROPOSAL_ERROR_INVALID_RISK_PARAMETER\x10\x1e\x12\x31\n-PROPOSAL_ERROR_MAJORITY_THRESHOLD_NOT_REACHED\x10\x1f\x12\x36\n2PROPOSAL_ERROR_PARTICIPATION_THRESHOLD_NOT_REACHED\x10 \x12(\n$PROPOSAL_ERROR_INVALID_ASSET_DETAILS\x10!\x12\x1f\n\x1bPROPOSAL_ERROR_UNKNOWN_TYPE\x10"\x12.\n*PROPOSAL_ERROR_UNKNOWN_RISK_PARAMETER_TYPE\x10#\x12#\n\x1fPROPOSAL_ERROR_INVALID_FREEFORM\x10$\x12\x31\n-PROPOSAL_ERROR_INSUFFICIENT_EQUITY_LIKE_SHARE\x10%\x12!\n\x1dPROPOSAL_ERROR_INVALID_MARKET\x10&\x12\x31\n-PROPOSAL_ERROR_TOO_MANY_MARKET_DECIMAL_PLACES\x10\'\x12\x35\n1PROPOSAL_ERROR_TOO_MANY_PRICE_MONITORING_TRIGGERS\x10(\x12/\n+PROPOSAL_ERROR_ERC20_ADDRESS_ALREADY_IN_USE\x10)B\'Z%code.vegaprotocol.io/vega/protos/vegab\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "vega.governance_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z%code.vegaprotocol.io/vega/protos/vega"
    _GOVERNANCEDATA_YESPARTYENTRY._options = None
    _GOVERNANCEDATA_YESPARTYENTRY._serialized_options = b"8\001"
    _GOVERNANCEDATA_NOPARTYENTRY._options = None
    _GOVERNANCEDATA_NOPARTYENTRY._serialized_options = b"8\001"
    _PROPOSALERROR._serialized_start = 5017
    _PROPOSALERROR._serialized_end = 6634
    _FUTUREPRODUCT._serialized_start = 112
    _FUTUREPRODUCT._serialized_end = 575
    _INSTRUMENTCONFIGURATION._serialized_start = 577
    _INSTRUMENTCONFIGURATION._serialized_end = 700
    _NEWMARKETCONFIGURATION._serialized_start = 703
    _NEWMARKETCONFIGURATION._serialized_end = 1248
    _NEWMARKET._serialized_start = 1250
    _NEWMARKET._serialized_end = 1317
    _UPDATEMARKET._serialized_start = 1319
    _UPDATEMARKET._serialized_end = 1421
    _UPDATEMARKETCONFIGURATION._serialized_start = 1424
    _UPDATEMARKETCONFIGURATION._serialized_end = 1883
    _UPDATEINSTRUMENTCONFIGURATION._serialized_start = 1885
    _UPDATEINSTRUMENTCONFIGURATION._serialized_end = 2000
    _UPDATEFUTUREPRODUCT._serialized_start = 2003
    _UPDATEFUTUREPRODUCT._serialized_end = 2429
    _UPDATENETWORKPARAMETER._serialized_start = 2431
    _UPDATENETWORKPARAMETER._serialized_end = 2505
    _NEWASSET._serialized_start = 2507
    _NEWASSET._serialized_end = 2563
    _UPDATEASSET._serialized_start = 2565
    _UPDATEASSET._serialized_end = 2657
    _NEWFREEFORM._serialized_start = 2659
    _NEWFREEFORM._serialized_end = 2672
    _PROPOSALTERMS._serialized_start = 2675
    _PROPOSALTERMS._serialized_end = 3203
    _PROPOSALRATIONALE._serialized_start = 3205
    _PROPOSALRATIONALE._serialized_end = 3292
    _GOVERNANCEDATA._serialized_start = 3295
    _GOVERNANCEDATA._serialized_end = 3685
    _GOVERNANCEDATA_YESPARTYENTRY._serialized_start = 3542
    _GOVERNANCEDATA_YESPARTYENTRY._serialized_end = 3613
    _GOVERNANCEDATA_NOPARTYENTRY._serialized_start = 3615
    _GOVERNANCEDATA_NOPARTYENTRY._serialized_end = 3685
    _PROPOSAL._serialized_start = 3688
    _PROPOSAL._serialized_end = 4610
    _PROPOSAL_STATE._serialized_start = 4320
    _PROPOSAL_STATE._serialized_end = 4494
    _VOTE._serialized_start = 4613
    _VOTE._serialized_end = 5014
    _VOTE_VALUE._serialized_start = 4955
    _VOTE_VALUE._serialized_end = 5014
# @@protoc_insertion_point(module_scope)



from concurrent.futures import ThreadPoolExecutor
from enum import Enum, auto
from io import BufferedWriter
from typing import Any
from logging import getLogger

import os
import dotenv
import numpy as np
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from typing import Optional
from vega_sim.grpc.client import VegaCoreClient
from vega_sim.wallet.base import Wallet

import vega_sim.proto.vega.api.v1.core_pb2 as core_proto
import vega_sim.proto.vega.commands.v1.transaction_pb2 as transaction_proto
import vega_sim.proto.vega.commands.v1.signature_pb2 as signature_proto
from vega_sim.wallet.vega_wallet import VegaWallet
from vega_sim.wallet.base import VEGA_DEFAULT_KEY_NAME

logger = getLogger(__name__)


class SlimWallet(Wallet):
    def __init__(
        self,
        core_client: VegaCoreClient,
        height_update_frequency: int = 500,
        full_wallet: Optional[VegaWallet] = None,
        full_wallet_default_pass: str = "passwd",
        store_transactions: bool = False,
        log_dir: Optional[str] = None,
    ):
        """Creates a wallet to running key generation internally
        and directly sending transactions to the Core node

        Args:
            core_client:
                VegaCoreClient, client for vega core interaction
            height_update_frequency:
                int, default 500, how frequently to update the block height parameter
                    for transactions (e.g. 500 will update every 500 calls to
                    submit_transaction).
            full_wallet:
                Optional[VegaWallet], optional full wallet backing. This will be used
                    for creating keypairs instead of generating them locally, making it
                    possible to connect to the Console
            full_wallet_default_pass:
                str, default 'passwd', If full wallet is passed, the password used
                    when creating dummy accounts if none are passed.
                    Use this password to log in to the Vega Console
            store_transactions:
                bool, default False, If True will store every transaction sent into
                    a file, allowing replay of the chain without going through full
                    logic of actors etc
        """
        self.core_client = core_client
        self.keys = {}
        self.pub_keys = {}
        self.idx = 0
        self.nonce_idx = 0
        self.pool = ThreadPoolExecutor(max_workers=5)

        self.height_update_frequency = 500
        self.remaining_until_height_update = 0
        self.block_height = None

        self.vega_wallet = full_wallet
        self.full_wallet_default_pass = full_wallet_default_pass

        self.store_transactions = store_transactions
        self.log_dir = log_dir

        # If it turns out that customising these is useful it's trivial to
        # make a parameter
        self.num_sigs_to_create = 1000000
        self.num_nonces_to_create = 1000000

        self._create_nonces()
        self._create_sigs()

        dotenv.load_dotenv()
        self.vega_default_key_name = os.environ.get(
            "VEGA_DEFAULT_KEY_NAME", VEGA_DEFAULT_KEY_NAME
        )

    def _create_sigs(self):
        self.sigs = [os.urandom(6).hex() for _ in range(self.num_sigs_to_create)]

    def _create_nonces(self):
        self.nonces = np.random.randint(0, 100000, size=self.num_nonces_to_create)

    def _next_sig(self) -> bytes:
        self.idx += 1
        if self.idx >= self.num_sigs_to_create:
            self.idx = 0
            self._create_sigs()
        return self.sigs[self.idx]

    def _next_nonce(self) -> int:
        self.nonce_idx += 1
        if self.nonce_idx >= self.num_nonces_to_create:
            self.nonce_idx = 0
            self._create_nonces()
        return self.nonces[self.nonce_idx]

    def create_wallet(
        self,
        name: str,
        passphrase: Optional[str] = None,
        key_name: Optional[str] = None,
    ) -> None:
        """Generates a new wallet from a name - passphrase pair in the given vega service.

        Args:
            name:
                str, The name to use for the wallet
        """
        passphrase = (
            passphrase if passphrase is not None else self.full_wallet_default_pass
        )
        key_name = key_name if key_name is not None else self.vega_default_key_name

        if name not in self.keys:
            self.keys[name] = {}
        if name not in self.pub_keys:
            self.pub_keys[name] = {}

        if self.vega_wallet is None:
            self.keys[name][key_name] = SigningKey.generate()
            self.pub_keys[name][key_name] = (
                self.keys[name][key_name].verify_key.encode(encoder=HexEncoder).decode()
            )

        else:
            self.vega_wallet.create_wallet(
                name=name,
                passphrase=passphrase,
                key_name=key_name,
            )
            self.pub_keys[name][key_name] = self.vega_wallet.public_key(name, key_name)

    def login(self, name: str, **kwargs) -> None:
        """Logs in to existing wallet in the given vega service.

        Args:
            name:
                str, The name of the wallet
        """
        if name not in self.keys:
            self.create_wallet(name=name)

    def submit_transaction(
        self,
        name: str,
        transaction: Any,
        transaction_type: str,
        key_name: Optional[str] = None,
    ):
        # if self.remaining_until_height_update <= 0:
        #     self.block_height = self.core_client.LastBlockHeight(
        #         core_proto.LastBlockHeightRequest()
        #     ).height
        #     self.remaining_until_height_update = self.height_update_frequency

        pub_key = self.public_key(name=name, key_name=key_name)

        transaction_info = {transaction_type: transaction}
        input_data = transaction_proto.InputData(
            nonce=self._next_nonce(),  # block_height=self.block_height,
            **transaction_info
        )

        serialised = input_data.SerializeToString()

        trans = transaction_proto.Transaction(
            input_data=serialised,
            signature=signature_proto.Signature(
                value=self._next_sig(),
                algo="vega/ed25519",
                version=1,
            ),
            pub_key=pub_key,
            version=3,
            pow=transaction_proto.ProofOfWork(
                tid=None,
                nonce=0,
            ),
        )
        request = core_proto.SubmitTransactionRequest(
            tx=trans, type=core_proto.SubmitTransactionRequest.Type.TYPE_ASYNC
        )

        submit_future = self.core_client.SubmitTransaction.future(request)
        self.pool.submit(lambda: submit_future.result())
        self.remaining_until_height_update -= 1

    def submit_raw_transaction(self, transaction: core_proto.SubmitTransactionRequest):
        submit_future = self.core_client.SubmitTransaction.future(transaction)
        self.pool.submit(lambda: submit_future.result())

    def public_key(self, name: str, key_name: Optional[str] = None) -> str:
        """Return a public key for the given wallet name and key name.

        Args:
            name (str):
                Name of the wallet.
            key_name (str):
                Name of the key. Defaults to None.

        Returns:
            str, public key
        """

        if key_name is None:
            return self.pub_keys[name][self.vega_default_key_name]
        else:
            return self.pub_keys[name][key_name]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/commands/v1/commands.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ... import governance_pb2 as vega_dot_governance__pb2
from ... import vega_pb2 as vega_dot_vega__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fvega/commands/v1/commands.proto\x12\x10vega.commands.v1\x1a\x15vega/governance.proto\x1a\x0fvega/vega.proto"\xeb\x01\n\x17\x42\x61tchMarketInstructions\x12I\n\rcancellations\x18\x01 \x03(\x0b\x32#.vega.commands.v1.OrderCancellationR\rcancellations\x12@\n\namendments\x18\x02 \x03(\x0b\x32 .vega.commands.v1.OrderAmendmentR\namendments\x12\x43\n\x0bsubmissions\x18\x03 \x03(\x0b\x32!.vega.commands.v1.OrderSubmissionR\x0bsubmissions"\xce\x02\n\x0fOrderSubmission\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12\x14\n\x05price\x18\x02 \x01(\tR\x05price\x12\x12\n\x04size\x18\x03 \x01(\x04R\x04size\x12\x1e\n\x04side\x18\x04 \x01(\x0e\x32\n.vega.SideR\x04side\x12;\n\rtime_in_force\x18\x05 \x01(\x0e\x32\x17.vega.Order.TimeInForceR\x0btimeInForce\x12\x1d\n\nexpires_at\x18\x06 \x01(\x03R\texpiresAt\x12$\n\x04type\x18\x07 \x01(\x0e\x32\x10.vega.Order.TypeR\x04type\x12\x1c\n\treference\x18\x08 \x01(\tR\treference\x12\x34\n\x0cpegged_order\x18\t \x01(\x0b\x32\x11.vega.PeggedOrderR\x0bpeggedOrder"K\n\x11OrderCancellation\x12\x19\n\x08order_id\x18\x01 \x01(\tR\x07orderId\x12\x1b\n\tmarket_id\x18\x02 \x01(\tR\x08marketId"\xe3\x02\n\x0eOrderAmendment\x12\x19\n\x08order_id\x18\x01 \x01(\tR\x07orderId\x12\x1b\n\tmarket_id\x18\x02 \x01(\tR\x08marketId\x12\x19\n\x05price\x18\x03 \x01(\tH\x00R\x05price\x88\x01\x01\x12\x1d\n\nsize_delta\x18\x04 \x01(\x03R\tsizeDelta\x12"\n\nexpires_at\x18\x05 \x01(\x03H\x01R\texpiresAt\x88\x01\x01\x12;\n\rtime_in_force\x18\x06 \x01(\x0e\x32\x17.vega.Order.TimeInForceR\x0btimeInForce\x12#\n\rpegged_offset\x18\x07 \x01(\tR\x0cpeggedOffset\x12@\n\x10pegged_reference\x18\x08 \x01(\x0e\x32\x15.vega.PeggedReferenceR\x0fpeggedReferenceB\x08\n\x06_priceB\r\n\x0b_expires_at"\xee\x01\n\x1cLiquidityProvisionSubmission\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12+\n\x11\x63ommitment_amount\x18\x02 \x01(\tR\x10\x63ommitmentAmount\x12\x10\n\x03\x66\x65\x65\x18\x03 \x01(\tR\x03\x66\x65\x65\x12*\n\x05sells\x18\x04 \x03(\x0b\x32\x14.vega.LiquidityOrderR\x05sells\x12(\n\x04\x62uys\x18\x05 \x03(\x0b\x32\x14.vega.LiquidityOrderR\x04\x62uys\x12\x1c\n\treference\x18\x06 \x01(\tR\treference"=\n\x1eLiquidityProvisionCancellation\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId"\xed\x01\n\x1bLiquidityProvisionAmendment\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12+\n\x11\x63ommitment_amount\x18\x02 \x01(\tR\x10\x63ommitmentAmount\x12\x10\n\x03\x66\x65\x65\x18\x03 \x01(\tR\x03\x66\x65\x65\x12*\n\x05sells\x18\x04 \x03(\x0b\x32\x14.vega.LiquidityOrderR\x05sells\x12(\n\x04\x62uys\x18\x05 \x03(\x0b\x32\x14.vega.LiquidityOrderR\x04\x62uys\x12\x1c\n\treference\x18\x06 \x01(\tR\treference"g\n\x12WithdrawSubmission\x12\x16\n\x06\x61mount\x18\x01 \x01(\tR\x06\x61mount\x12\x14\n\x05\x61sset\x18\x02 \x01(\tR\x05\x61sset\x12#\n\x03\x65xt\x18\x03 \x01(\x0b\x32\x11.vega.WithdrawExtR\x03\x65xt"\x94\x01\n\x12ProposalSubmission\x12\x1c\n\treference\x18\x01 \x01(\tR\treference\x12)\n\x05terms\x18\x02 \x01(\x0b\x32\x13.vega.ProposalTermsR\x05terms\x12\x35\n\trationale\x18\x03 \x01(\x0b\x32\x17.vega.ProposalRationaleR\trationale"Y\n\x0eVoteSubmission\x12\x1f\n\x0bproposal_id\x18\x01 \x01(\tR\nproposalId\x12&\n\x05value\x18\x02 \x01(\x0e\x32\x10.vega.Vote.ValueR\x05value"E\n\x12\x44\x65legateSubmission\x12\x17\n\x07node_id\x18\x01 \x01(\tR\x06nodeId\x12\x16\n\x06\x61mount\x18\x02 \x01(\tR\x06\x61mount"\xf1\x01\n\x14UndelegateSubmission\x12\x17\n\x07node_id\x18\x01 \x01(\tR\x06nodeId\x12\x16\n\x06\x61mount\x18\x02 \x01(\tR\x06\x61mount\x12\x45\n\x06method\x18\x03 \x01(\x0e\x32-.vega.commands.v1.UndelegateSubmission.MethodR\x06method"a\n\x06Method\x12\x16\n\x12METHOD_UNSPECIFIED\x10\x00\x12\x0e\n\nMETHOD_NOW\x10\x01\x12\x1a\n\x16METHOD_AT_END_OF_EPOCH\x10\x02\x12\x13\n\x0fMETHOD_IN_ANGER\x10\x03"\xea\x02\n\x08Transfer\x12=\n\x11\x66rom_account_type\x18\x01 \x01(\x0e\x32\x11.vega.AccountTypeR\x0f\x66romAccountType\x12\x0e\n\x02to\x18\x02 \x01(\tR\x02to\x12\x39\n\x0fto_account_type\x18\x03 \x01(\x0e\x32\x11.vega.AccountTypeR\rtoAccountType\x12\x14\n\x05\x61sset\x18\x04 \x01(\tR\x05\x61sset\x12\x16\n\x06\x61mount\x18\x05 \x01(\tR\x06\x61mount\x12\x1c\n\treference\x18\x06 \x01(\tR\treference\x12;\n\x07one_off\x18\x65 \x01(\x0b\x32 .vega.commands.v1.OneOffTransferH\x00R\x06oneOff\x12\x43\n\trecurring\x18\x66 \x01(\x0b\x32#.vega.commands.v1.RecurringTransferH\x00R\trecurringB\x06\n\x04kind"/\n\x0eOneOffTransfer\x12\x1d\n\ndeliver_on\x18\x01 \x01(\x03R\tdeliverOn"\xc1\x01\n\x11RecurringTransfer\x12\x1f\n\x0bstart_epoch\x18\x01 \x01(\x04R\nstartEpoch\x12 \n\tend_epoch\x18\x02 \x01(\x04H\x00R\x08\x65ndEpoch\x88\x01\x01\x12\x16\n\x06\x66\x61\x63tor\x18\x03 \x01(\tR\x06\x66\x61\x63tor\x12\x43\n\x11\x64ispatch_strategy\x18\x04 \x01(\x0b\x32\x16.vega.DispatchStrategyR\x10\x64ispatchStrategyB\x0c\n\n_end_epoch"1\n\x0e\x43\x61ncelTransfer\x12\x1f\n\x0btransfer_id\x18\x01 \x01(\tR\ntransferIdB3Z1code.vegaprotocol.io/vega/protos/vega/commands/v1b\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "vega.commands.v1.commands_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z1code.vegaprotocol.io/vega/protos/vega/commands/v1"
    )
    _BATCHMARKETINSTRUCTIONS._serialized_start = 94
    _BATCHMARKETINSTRUCTIONS._serialized_end = 329
    _ORDERSUBMISSION._serialized_start = 332
    _ORDERSUBMISSION._serialized_end = 666
    _ORDERCANCELLATION._serialized_start = 668
    _ORDERCANCELLATION._serialized_end = 743
    _ORDERAMENDMENT._serialized_start = 746
    _ORDERAMENDMENT._serialized_end = 1101
    _LIQUIDITYPROVISIONSUBMISSION._serialized_start = 1104
    _LIQUIDITYPROVISIONSUBMISSION._serialized_end = 1342
    _LIQUIDITYPROVISIONCANCELLATION._serialized_start = 1344
    _LIQUIDITYPROVISIONCANCELLATION._serialized_end = 1405
    _LIQUIDITYPROVISIONAMENDMENT._serialized_start = 1408
    _LIQUIDITYPROVISIONAMENDMENT._serialized_end = 1645
    _WITHDRAWSUBMISSION._serialized_start = 1647
    _WITHDRAWSUBMISSION._serialized_end = 1750
    _PROPOSALSUBMISSION._serialized_start = 1753
    _PROPOSALSUBMISSION._serialized_end = 1901
    _VOTESUBMISSION._serialized_start = 1903
    _VOTESUBMISSION._serialized_end = 1992
    _DELEGATESUBMISSION._serialized_start = 1994
    _DELEGATESUBMISSION._serialized_end = 2063
    _UNDELEGATESUBMISSION._serialized_start = 2066
    _UNDELEGATESUBMISSION._serialized_end = 2307
    _UNDELEGATESUBMISSION_METHOD._serialized_start = 2210
    _UNDELEGATESUBMISSION_METHOD._serialized_end = 2307
    _TRANSFER._serialized_start = 2310
    _TRANSFER._serialized_end = 2672
    _ONEOFFTRANSFER._serialized_start = 2674
    _ONEOFFTRANSFER._serialized_end = 2721
    _RECURRINGTRANSFER._serialized_start = 2724
    _RECURRINGTRANSFER._serialized_end = 2917
    _CANCELTRANSFER._serialized_start = 2919
    _CANCELTRANSFER._serialized_end = 2968
# @@protoc_insertion_point(module_scope)



"""vega_load_test.scenario.py

Module contains a VegaLoadTest class for setting up a vega-market-sim scenario capable
of load testing the vega node and data nodes as well as generating data across long
periods.The number of parties, orders per second, and trades per second can be 
configured to vary the load on the vega node and date node.

A VegaLoadTest scenario contains the following agents for each market:

    • MarketManager - propose and settle markets
    • ExponentialShapedMarketMaker - provide order book depth
    • OpenAuctionPass - provide limit orders to pass opening auction
    • LimitOrderTrader - provide limit orders to meet orders per second quota
    • MarketOrderTrader - provide market orders to meet trades per second quota

Arguments:

    Currently market makers are optimised for the hard-coded default market arguments 
    and data sources which are configured by the following module level arguments.

    • START_DATE: start date for price data source
    • MARKET_A_ARGS: dictionary defining market a parameters
    • MARKET_B_ARGS: dictionary defining market b parameters
    • MARKET_C_ARGS: dictionary defining market c parameters


Examples:

    Runs a standard a preconfigured load test scenario with realistic market traffic
    parameters. The market will appear in opening auction for a minute whilst the agents
    are initiated.

    $ python -m vega_sim.scenario.adhoc -s vega_load_test --console --pause --debug

"""

import argparse
import logging
import numpy as np
from datetime import datetime, timedelta
from typing import Optional, Callable
from vega_sim.scenario.common.utils.price_process import (
    Granularity,
    get_historic_price_series,
)

from vega_sim.scenario.scenario import Scenario
from vega_sim.environment.environment import MarketEnvironmentWithState
from vega_sim.null_service import VegaServiceNull
from vega_sim.scenario.constants import Network
from vega_sim.scenario.multi_market.agents import (
    MARKET_MANAGERS,
    MARKET_MAKERS,
    MARKET_PASSERS,
    MARKET_TRADERS,
)
from vega_sim.scenario.common.agents import (
    MarketManager,
    OpenAuctionPass,
    ExponentialShapedMarketMaker,
    MarketOrderTrader,
    LimitOrderTrader,
)

START_DATE = "2021-06-01 00:00:00"

MARKET_A_ARGS = {
    "name": "ETH:USD (12 Month)",
    "restarts": 1,
    "oracle": "ETH-USD",
    "asset": "tUSD",
    "mdp": 5,
    "pdp": 2,
    "adp": 5,
}

MARKET_B_ARGS = {
    "name": "ETH:GBP (6 Month)",
    "restarts": 2,
    "oracle": "ETH-GBP",
    "asset": "tGBP",
    "mdp": 5,
    "pdp": 2,
    "adp": 5,
}

MARKET_C_ARGS = {
    "name": "BTC:EUR (1 Month)",
    "restarts": 12,
    "oracle": "BTC-EUR",
    "asset": "tEUR",
    "mdp": 5,
    "pdp": 2,
    "adp": 5,
}


class VegaLoadTest(Scenario):
    def __init__(
        self,
        num_steps: int = 60 * 3,
        granularity: Granularity = Granularity.MINUTE,
        transactions_per_block: int = 4096,
        block_length_seconds: float = 1,
        parties_per_market: int = 1000,
        orders_per_second: int = 100,
        trades_per_second: int = 1,
        start_date: str = None,
        market_a_args: Optional[dict] = None,
        market_b_args: Optional[dict] = None,
        market_c_args: Optional[dict] = None,
        initial_asset_mint=1e9,
        price_process_fn: Optional[Callable] = None,
    ):

        self.num_steps = num_steps
        self.granularity = granularity

        self.block_length_seconds = block_length_seconds
        self.transactions_per_block = transactions_per_block

        self.start_date = start_date if start_date is not None else START_DATE

        self.market_a_args = (
            market_a_args if market_a_args is not None else MARKET_A_ARGS
        )
        self.market_b_args = (
            market_b_args if market_b_args is not None else MARKET_B_ARGS
        )
        self.market_c_args = (
            market_c_args if market_c_args is not None else MARKET_C_ARGS
        )

        self.initial_asset_mint = initial_asset_mint

        self.num_lo_traders_per_market = min(
            [orders_per_second, parties_per_market - 1]
        )
        self.num_mo_traders_per_market = (
            parties_per_market - self.num_lo_traders_per_market
        )

        self.market_order_trader_step_bias = (
            trades_per_second / self.num_mo_traders_per_market
        ) * self.granularity.value

        self.price_process_fn = price_process_fn

    def _generate_price_process(self, asset: str) -> list:

        start = datetime.strptime(self.start_date, "%Y-%m-%d %H:%M:%S")
        end = start + timedelta(seconds=self.num_steps * self.granularity.value)

        price_process = get_historic_price_series(
            product_id=asset,
            granularity=self.granularity,
            start=str(start),
            end=str(end),
        )

        return list(price_process)

    def set_up_background_market(
        self,
        vega: VegaServiceNull,
        random_state: Optional[np.random.RandomState] = None,
    ) -> MarketEnvironmentWithState:

        market_a_price_process = (
            self.price_process_fn()
            if self.price_process_fn is not None
            else self._generate_price_process(asset=self.market_a_args["oracle"])
        )
        market_b_price_process = (
            self.price_process_fn()
            if self.price_process_fn is not None
            else self._generate_price_process(asset=self.market_a_args["oracle"])
        )
        market_c_price_process = (
            self.price_process_fn()
            if self.price_process_fn is not None
            else self._generate_price_process(asset=self.market_a_args["oracle"])
        )

        # Create MarketManager agents
        market_a_manager = MarketManager(
            wallet_name=MARKET_MANAGERS["MARKET_A_CREATOR"].wallet_name,
            wallet_pass=MARKET_MANAGERS["MARKET_A_CREATOR"].wallet_name,
            key_name=MARKET_MANAGERS["MARKET_A_CREATOR"].key_name,
            terminate_wallet_name=MARKET_MANAGERS["MARKET_A_SETTLER"].wallet_name,
            terminate_wallet_pass=MARKET_MANAGERS["MARKET_A_SETTLER"].wallet_name,
            terminate_key_name=MARKET_MANAGERS["MARKET_A_SETTLER"].key_name,
            market_name=self.market_a_args["name"],
            asset_name=self.market_a_args["asset"],
            asset_decimal=self.market_a_args["adp"],
            market_decimal=self.market_a_args["mdp"],
            market_position_decimal=self.market_a_args["pdp"],
            settlement_price=market_a_price_process[-1],
        )
        market_b_manager = MarketManager(
            wallet_name=MARKET_MANAGERS["MARKET_B_CREATOR"].wallet_name,
            wallet_pass=MARKET_MANAGERS["MARKET_B_CREATOR"].wallet_name,
            key_name=MARKET_MANAGERS["MARKET_B_CREATOR"].key_name,
            terminate_wallet_name=MARKET_MANAGERS["MARKET_B_SETTLER"].wallet_name,
            terminate_wallet_pass=MARKET_MANAGERS["MARKET_B_SETTLER"].wallet_name,
            terminate_key_name=MARKET_MANAGERS["MARKET_B_SETTLER"].key_name,
            market_name=self.market_b_args["name"],
            asset_name=self.market_b_args["asset"],
            asset_decimal=self.market_b_args["adp"],
            market_decimal=self.market_b_args["mdp"],
            market_position_decimal=self.market_b_args["pdp"],
            settlement_price=market_b_price_process[-1],
        )
        market_c_manager = MarketManager(
            wallet_name=MARKET_MANAGERS["MARKET_B_CREATOR"].wallet_name,
            wallet_pass=MARKET_MANAGERS["MARKET_B_CREATOR"].wallet_name,
            key_name=MARKET_MANAGERS["MARKET_B_CREATOR"].key_name,
            terminate_wallet_name=MARKET_MANAGERS["MARKET_C_SETTLER"].wallet_name,
            terminate_wallet_pass=MARKET_MANAGERS["MARKET_C_SETTLER"].wallet_name,
            terminate_key_name=MARKET_MANAGERS["MARKET_C_SETTLER"].key_name,
            market_name=self.market_c_args["name"],
            asset_name=self.market_c_args["asset"],
            asset_decimal=self.market_c_args["adp"],
            market_decimal=self.market_c_args["mdp"],
            market_position_decimal=self.market_c_args["pdp"],
            settlement_price=market_c_price_process[-1],
        )

        #  Create ExponentialShapedMarketMaker agents
        market_a_maker = ExponentialShapedMarketMaker(
            wallet_name=MARKET_MAKERS["MARKET_A_MAKER"].wallet_name,
            wallet_pass=MARKET_MAKERS["MARKET_A_MAKER"].wallet_pass,
            key_name=MARKET_MAKERS["MARKET_A_MAKER"].key_name,
            price_process_generator=iter(market_a_price_process),
            initial_asset_mint=self.initial_asset_mint,
            market_name=self.market_a_args["name"],
            asset_name=self.market_a_args["asset"],
            commitment_amount=1e9,
            market_decimal_places=self.market_a_args["mdp"],
            asset_decimal_places=self.market_a_args["adp"],
            num_steps=self.num_steps,
            tick_spacing=1,
            market_kappa=10,
        )
        market_b_maker = ExponentialShapedMarketMaker(
            wallet_name=MARKET_MAKERS["MARKET_B_MAKER"].wallet_name,
            wallet_pass=MARKET_MAKERS["MARKET_B_MAKER"].wallet_pass,
            key_name=MARKET_MAKERS["MARKET_B_MAKER"].key_name,
            price_process_generator=iter(market_b_price_process),
            initial_asset_mint=self.initial_asset_mint,
            market_name=self.market_b_args["name"],
            asset_name=self.market_b_args["asset"],
            commitment_amount=1e9,
            market_decimal_places=self.market_b_args["mdp"],
            asset_decimal_places=self.market_b_args["adp"],
            num_steps=self.num_steps,
            tick_spacing=1,
            market_kappa=10,
        )
        market_c_maker = ExponentialShapedMarketMaker(
            wallet_name=MARKET_MAKERS["MARKET_C_MAKER"].wallet_name,
            wallet_pass=MARKET_MAKERS["MARKET_C_MAKER"].wallet_pass,
            key_name=MARKET_MAKERS["MARKET_C_MAKER"].key_name,
            price_process_generator=iter(market_c_price_process),
            initial_asset_mint=self.initial_asset_mint,
            market_name=self.market_c_args["name"],
            asset_name=self.market_c_args["asset"],
            commitment_amount=1e9,
            market_decimal_places=self.market_c_args["mdp"],
            asset_decimal_places=self.market_c_args["adp"],
            num_steps=self.num_steps,
            tick_spacing=2,
            market_kappa=5,
        )

        # Setup agents for passing auction
        market_a_passer_bid = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_A_PASSER_BID"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_A_PASSER_BID"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_A_PASSER_BID"].key_name,
            side="SIDE_BUY",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_a_price_process[0],
            market_name=self.market_a_args["name"],
            asset_name=self.market_a_args["asset"],
            opening_auction_trade_amount=1,
        )
        market_a_passer_ask = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_A_PASSER_ASK"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_A_PASSER_ASK"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_A_PASSER_ASK"].key_name,
            side="SIDE_SELL",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_a_price_process[0],
            market_name=self.market_a_args["name"],
            asset_name=self.market_a_args["asset"],
            opening_auction_trade_amount=1,
        )
        market_b_passer_bid = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_B_PASSER_BID"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_B_PASSER_BID"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_B_PASSER_BID"].key_name,
            side="SIDE_BUY",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_b_price_process[0],
            market_name=self.market_b_args["name"],
            asset_name=self.market_b_args["asset"],
            opening_auction_trade_amount=1,
        )
        market_b_passer_ask = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_B_PASSER_ASK"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_B_PASSER_ASK"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_B_PASSER_ASK"].key_name,
            side="SIDE_SELL",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_b_price_process[0],
            market_name=self.market_b_args["name"],
            asset_name=self.market_b_args["asset"],
            opening_auction_trade_amount=1,
        )
        market_c_passer_bid = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_C_PASSER_BID"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_C_PASSER_BID"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_C_PASSER_BID"].key_name,
            side="SIDE_BUY",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_c_price_process[0],
            market_name=self.market_c_args["name"],
            asset_name=self.market_c_args["asset"],
            opening_auction_trade_amount=1,
        )
        market_c_passer_ask = OpenAuctionPass(
            wallet_name=MARKET_PASSERS["MARKET_C_PASSER_ASK"].wallet_name,
            wallet_pass=MARKET_PASSERS["MARKET_C_PASSER_ASK"].wallet_pass,
            key_name=MARKET_PASSERS["MARKET_C_PASSER_ASK"].key_name,
            side="SIDE_SELL",
            initial_asset_mint=self.initial_asset_mint,
            initial_price=market_c_price_process[0],
            market_name=self.market_c_args["name"],
            asset_name=self.market_c_args["asset"],
            opening_auction_trade_amount=1,
        )

        market_a_lo_traders = [
            LimitOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_A_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_a_args["name"],
                asset_name=self.market_a_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                buy_volume=1,
                sell_volume=1,
                submit_bias=1,
                cancel_bias=1,
            )
            for i in range(self.num_lo_traders_per_market)
        ]

        market_b_lo_traders = [
            LimitOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_A_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_a_args["name"],
                asset_name=self.market_a_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                buy_volume=1,
                sell_volume=1,
                submit_bias=1,
                cancel_bias=1,
            )
            for i in range(self.num_lo_traders_per_market)
        ]

        market_c_lo_traders = [
            LimitOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_A_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_a_args["name"],
                asset_name=self.market_a_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                buy_volume=1,
                sell_volume=1,
                submit_bias=1,
                cancel_bias=1,
            )
            for i in range(self.num_lo_traders_per_market)
        ]

        market_a_mo_traders = [
            MarketOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_A_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_A_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_a_args["name"],
                asset_name=self.market_a_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                base_order_size=1,
                step_bias=self.market_order_trader_step_bias,
            )
            for i in range(self.num_mo_traders_per_market)
        ]
        market_b_mo_traders = [
            MarketOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_B_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_B_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_B_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_b_args["name"],
                asset_name=self.market_b_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                base_order_size=1,
                step_bias=self.market_order_trader_step_bias,
            )
            for i in range(self.num_mo_traders_per_market)
        ]
        market_c_mo_traders = [
            MarketOrderTrader(
                wallet_name=MARKET_TRADERS[
                    f"MARKET_C_TRADER_{str(i).zfill(4)}"
                ].wallet_name,
                wallet_pass=MARKET_TRADERS[
                    f"MARKET_C_TRADER_{str(i).zfill(4)}"
                ].wallet_pass,
                key_name=MARKET_TRADERS[f"MARKET_C_TRADER_{str(i).zfill(4)}"].key_name,
                market_name=self.market_c_args["name"],
                asset_name=self.market_c_args["asset"],
                buy_intensity=10,
                sell_intensity=10,
                base_order_size=1,
                step_bias=self.market_order_trader_step_bias,
            )
            for i in range(self.num_mo_traders_per_market)
        ]

        env = MarketEnvironmentWithState(
            agents=[
                market_a_manager,
                market_b_manager,
                market_c_manager,
                market_a_passer_bid,
                market_a_passer_ask,
                market_b_passer_bid,
                market_b_passer_ask,
                market_c_passer_bid,
                market_c_passer_ask,
                market_a_maker,
                market_b_maker,
                market_c_maker,
            ]
            + market_a_lo_traders
            + market_b_lo_traders
            + market_c_lo_traders
            + market_a_mo_traders
            + market_b_mo_traders
            + market_c_mo_traders,
            n_steps=self.num_steps,
            random_agent_ordering=False,
            transactions_per_block=self.transactions_per_block,
            vega_service=vega,
            step_length_seconds=self.granularity.value,
            block_length_seconds=vega.seconds_per_block,
        )
        return env

    def run_iteration(
        self,
        vega: VegaServiceNull,
        network: Optional[Network] = None,
        pause_at_completion: bool = False,
        run_with_console: bool = False,
        random_state: Optional[np.random.RandomState] = None,
    ):
        env = self.set_up_background_market(vega=vega, random_state=random_state)
        result = env.run(
            pause_at_completion=pause_at_completion,
            run_with_console=run_with_console,
        )
        return result


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    scenario = VegaLoadTest(
        num_steps=1000,
        granularity=Granularity.FIFTEEN_MINUTE,
        block_length_seconds=60,
        transactions_per_block=4096,
    )

    with VegaServiceNull(
        warn_on_raw_data_access=False,
        run_with_console=True,
        use_full_vega_wallet=False,
        retain_log_files=True,
        launch_graphql=False,
        seconds_per_block=scenario.block_length_seconds,
        transactions_per_block=scenario.transactions_per_block,
    ) as vega:
        scenario.run_iteration(
            vega=vega,
            pause_at_completion=False,
        )



from dataclasses import dataclass
from typing import List, Tuple
import numpy as np

from vega_sim.environment.agent import Agent


@dataclass
class AbstractAction:
    pass


@dataclass
class LAMarketState:
    step: int
    position: float
    full_balance: float
    market_in_auction: bool
    bid_prices: List[float]
    ask_prices: List[float]
    bid_volumes: List[float]
    ask_volumes: List[float]
    trading_fee: float
    next_price: float

    def to_array(self):
        l = (
            [
                self.step,
                self.position,
                self.full_balance,
                int(self.market_in_auction),
                self.trading_fee,
                self.next_price,
            ]
            + self.bid_prices
            + self.ask_prices
            + self.bid_volumes
            + self.ask_volumes
        )
        arr = np.nan_to_num(np.array(l))
        norm_factor = max(1e-12, np.linalg.norm(arr))
        arr = arr / norm_factor
        return arr

    @staticmethod
    def from_array(arr: np.array, num_lvls: int):
        step = arr[0]
        pos = arr[1]
        balance = arr[2]
        auction = arr[3]
        fees = arr[4]
        next_p = arr[5]
        offset = 6
        bid_prices = arr[offset : (offset + num_lvls)]
        offset = offset + num_lvls + 1
        ask_prices = arr[offset : (offset + num_lvls)]
        offset = offset + num_lvls + 1
        bid_volumes = arr[offset : (offset + num_lvls)]
        offset = offset + num_lvls + 1
        ask_volumes = arr[offset : (offset + num_lvls)]
        return LAMarketState(
            step=step,
            position=pos,
            full_balance=balance,
            auction=auction,
            fees=fees,
            next_price=next_p,
            bid_prices=bid_prices,
            ask_prices=ask_prices,
            bid_volumes=bid_volumes,
            ask_volumes=ask_volumes,
        )


def states_to_sarsa(
    states: List[Tuple[LAMarketState, AbstractAction]],
    inventory_penalty: float = 0.0,
) -> List[Tuple[LAMarketState, AbstractAction, float, LAMarketState, AbstractAction]]:
    res = []
    for i in range(len(states) - 1):
        pres_state = states[i]
        next_state = states[i + 1]

        if next_state[0].full_balance <= 0:
            reward = -1e12
            res.append(
                (
                    pres_state[0],
                    pres_state[1],
                    reward,
                    next_state[0] if next_state is not np.nan else np.nan,
                    next_state[1] if next_state is not np.nan else np.nan,
                )
            )
            break

        reward = (
            next_state[0].full_balance - pres_state[0].full_balance
            if next_state is not np.nan
            else 0
        )
        reward -= inventory_penalty * pres_state[0].position * pres_state[0].position
        res.append((pres_state[0], pres_state[1], reward, next_state[0], next_state[1]))
    return res



from abc import ABC
from dataclasses import dataclass
from typing import Any, Tuple, Dict, Optional

from vega_sim.service import VegaService


@dataclass
class VegaState:
    network_state: Tuple
    market_state: Dict[str, Any]


class Agent(ABC):
    def step(self, vega: VegaService):
        pass

    def initialise(
        self, vega: VegaService, create_wallet: bool = False, mint_wallet: bool = False
    ):
        self.vega = vega

    def finalise(self):
        pass


class AgentWithWallet(Agent):
    def __init__(
        self,
        wallet_name: str,
        wallet_pass: str,
        key_name: Optional[str] = None,
    ):
        """Agent for use in environments as specified in environment.py.
        To extend, the crucial function to implement is the step function which will
        be called on each timestep in the simulation.

        Additionally, the initialise function can be added to. This function is called
        once before the main simulation and can be used to creat assets, set up markets,
        faucet assets to the agent etc.

        Args:
            wallet_name:
                str, The name to use for this agent's wallet
            wallet_pass:
                str, The password which this agent uses to log in to the wallet
            key_name:
                str, optional, Name of key in wallet for agent to use. Defaults
                to value in the environment variable "VEGA_DEFAULT_KEY_NAME".
        """
        super().__init__()
        self.wallet_name = wallet_name
        self.wallet_pass = wallet_pass
        self.key_name = key_name

    def step(self, vega: VegaService):
        pass

    def initialise(self, vega: VegaService, create_wallet: bool = True):
        super().initialise(vega=vega)
        if create_wallet:
            self.vega.create_wallet(
                name=self.wallet_name,
                passphrase=self.wallet_pass,
                key_name=self.key_name,
            )
        else:
            self.vega.login(name=self.wallet_name, passphrase=self.wallet_pass)


class StateAgentWithWallet(AgentWithWallet):
    def step(self, vega_state: VegaState):
        pass


class StateAgent(Agent):
    def step(self, vega_state: VegaState):
        pass



import numpy as np

from datetime import datetime, timedelta
from typing import Any, Callable, List, Optional

from vega_sim.api.market import MarketConfig
from vega_sim.environment.agent import Agent
from vega_sim.scenario.common.utils.price_process import (
    Granularity,
    get_historic_price_series,
)

from vega_sim.scenario.scenario import Scenario
from vega_sim.environment.environment import MarketEnvironmentWithState
from vega_sim.null_service import VegaServiceNull
from vega_sim.scenario.constants import Network

from vega_sim.scenario.configurable_market.agents import (
    ConfigurableMarketManager,
    PROPOSAL_PARTY,
    TERMINATION_PARTY,
    MAKER_PARTY,
    AUCTION_PARTY_A,
    AUCTION_PARTY_B,
    SENSITIVE_PARTY_A,
    SENSITIVE_PARTY_B,
    SENSITIVE_PARTY_C,
    INFORMED_PARTY,
)

from vega_sim.scenario.common.agents import (
    OpenAuctionPass,
    ExponentialShapedMarketMaker,
    PriceSensitiveMarketOrderTrader,
    InformedTrader,
)


class ConfigurableMarket(Scenario):
    def __init__(
        self,
        market_name: str = None,
        market_code: str = None,
        asset_name: str = None,
        asset_dp: str = None,
        num_steps: int = 120,
        granularity: Optional[Granularity] = Granularity.MINUTE,
        block_size: int = 1,
        block_length_seconds: int = 1,
        state_extraction_freq: int = 1,
        state_extraction_fn: Optional[
            Callable[[VegaServiceNull, List[Agent]], Any]
        ] = None,
        settle_at_end: bool = True,
        price_process_fn: Optional[Callable] = None,
    ):

        # Simulation settings
        self.num_steps = num_steps
        self.granularity = granularity
        self.block_size = block_size
        self.block_length_seconds = block_length_seconds
        self.settle_at_end = settle_at_end
        self.price_process_fn = price_process_fn

        # Logging options
        self.state_extraction_freq = state_extraction_freq
        self.state_extraction_fn = state_extraction_fn

        # Asset parameters
        self.asset_name = asset_name
        self.asset_decimal = asset_dp

        # Market parameters
        self.market_name = market_name
        self.market_code = market_code

    def _generate_price_process(
        self,
        random_state,
    ) -> list:

        start = "2021-07-01 00:00:00"

        start = datetime.strptime(start, "%Y-%m-%d %H:%M:%S")

        start = start + timedelta(days=int(random_state.choice(range(90))))

        end = start + timedelta(seconds=self.num_steps * self.granularity.value)

        price_process = get_historic_price_series(
            product_id="ETH-USD",
            granularity=self.granularity,
            start=str(start),
            end=str(end),
        )

        return list(price_process)

    def set_up_background_market(
        self,
        vega: VegaServiceNull,
        market_config: Optional[MarketConfig] = None,
        random_state: Optional[np.random.RandomState] = None,
    ) -> MarketEnvironmentWithState:

        market_config = market_config if market_config is not None else MarketConfig()

        random_state = (
            random_state if random_state is not None else np.random.RandomState()
        )

        market_name = self.market_name
        asset_name = self.asset_name

        price_process = (
            self.price_process_fn()
            if self.price_process_fn is not None
            else self._generate_price_process(random_state=random_state)
        )

        market_manager = ConfigurableMarketManager(
            proposal_wallet_name=PROPOSAL_PARTY.wallet_name,
            proposal_wallet_pass=PROPOSAL_PARTY.wallet_pass,
            proposal_key_name=PROPOSAL_PARTY.key_name,
            termination_wallet_name=TERMINATION_PARTY.wallet_name,
            termination_wallet_pass=TERMINATION_PARTY.wallet_pass,
            termination_key_name=TERMINATION_PARTY.key_name,
            market_config=market_config,
            market_name=market_name,
            market_code=self.market_code,
            asset_dp=self.asset_decimal,
            asset_name=asset_name,
            settlement_price=price_process[-1] if self.settle_at_end else None,
        )

        shaped_mm = ExponentialShapedMarketMaker(
            wallet_name=MAKER_PARTY.wallet_name,
            wallet_pass=MAKER_PARTY.wallet_pass,
            key_name=MAKER_PARTY.key_name,
            price_process_generator=iter(price_process),
            initial_asset_mint=1e9,
            commitment_amount=1e6,
            market_name=market_name,
            asset_name=asset_name,
            market_decimal_places=market_config.decimal_places,
            asset_decimal_places=self.asset_decimal,
            num_steps=self.num_steps,
            kappa=1,
            num_levels=25,
            tick_spacing=1,
            order_unit_size=2,
            max_order_size=1000,
            inventory_upper_boundary=200,
            inventory_lower_boundary=-200,
            market_kappa=1,
            market_order_arrival_rate=10,
        )

        sensitive_mo_trader_a = PriceSensitiveMarketOrderTrader(
            wallet_name=SENSITIVE_PARTY_A.wallet_name,
            wallet_pass=SENSITIVE_PARTY_A.wallet_pass,
            key_name=SENSITIVE_PARTY_A.key_name,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=1e10,
            buy_intensity=1,
            sell_intensity=1,
            price_half_life=10,
            price_process_generator=iter(price_process),
            base_order_size=1,
        )

        sensitive_mo_trader_b = PriceSensitiveMarketOrderTrader(
            wallet_name=SENSITIVE_PARTY_B.wallet_name,
            wallet_pass=SENSITIVE_PARTY_B.wallet_pass,
            key_name=SENSITIVE_PARTY_B.key_name,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=1e10,
            buy_intensity=10,
            sell_intensity=10,
            price_half_life=1,
            price_process_generator=iter(price_process),
            base_order_size=1,
        )

        sensitive_mo_trader_c = PriceSensitiveMarketOrderTrader(
            wallet_name=SENSITIVE_PARTY_C.wallet_name,
            wallet_pass=SENSITIVE_PARTY_C.wallet_pass,
            key_name=SENSITIVE_PARTY_C.key_name,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=1e10,
            buy_intensity=100,
            sell_intensity=100,
            price_half_life=0.1,
            price_process_generator=iter(price_process),
            base_order_size=1,
        )

        auctionpass1 = OpenAuctionPass(
            wallet_name=AUCTION_PARTY_A.wallet_name,
            wallet_pass=AUCTION_PARTY_A.wallet_pass,
            key_name=AUCTION_PARTY_A.key_name,
            side="SIDE_BUY",
            initial_asset_mint=1e9,
            initial_price=price_process[0],
            market_name=market_name,
            asset_name=asset_name,
            opening_auction_trade_amount=1,
        )

        auctionpass2 = OpenAuctionPass(
            wallet_name=AUCTION_PARTY_B.wallet_name,
            wallet_pass=AUCTION_PARTY_B.wallet_pass,
            key_name=AUCTION_PARTY_B.key_name,
            side="SIDE_SELL",
            initial_asset_mint=1e9,
            initial_price=price_process[0],
            market_name=market_name,
            asset_name=asset_name,
            opening_auction_trade_amount=1,
        )

        info_trader = InformedTrader(
            wallet_name=INFORMED_PARTY.wallet_name,
            wallet_pass=INFORMED_PARTY.wallet_pass,
            key_name=INFORMED_PARTY.key_name,
            price_process=price_process,
            market_name=market_name,
            asset_name=asset_name,
            initial_asset_mint=1e10,
            proportion_taken=0.1,
        )

        env = MarketEnvironmentWithState(
            agents=[
                market_manager,
                shaped_mm,
                sensitive_mo_trader_a,
                sensitive_mo_trader_b,
                sensitive_mo_trader_c,
                auctionpass1,
                auctionpass2,
                info_trader,
            ],
            n_steps=self.num_steps,
            step_length_seconds=self.granularity.value,
            random_agent_ordering=True,
            transactions_per_block=self.block_size,
            vega_service=vega,
            state_extraction_freq=self.state_extraction_freq,
            block_length_seconds=self.block_length_seconds,
            state_extraction_fn=self.state_extraction_fn,
        )
        return env

    def run_iteration(
        self,
        vega: VegaServiceNull,
        network: Optional[Network] = None,
        pause_at_completion: bool = False,
        run_with_console: bool = False,
        random_state: Optional[np.random.RandomState] = None,
        market_config: Optional[MarketConfig] = None,
    ):
        env = self.set_up_background_market(
            vega=vega, random_state=random_state, market_config=market_config
        )
        result = env.run(
            pause_at_completion=pause_at_completion,
            run_with_console=run_with_console,
        )
        return result



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/api/v1/corestate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ... import assets_pb2 as vega_dot_assets__pb2
from ... import governance_pb2 as vega_dot_governance__pb2
from ... import markets_pb2 as vega_dot_markets__pb2
from ... import vega_pb2 as vega_dot_vega__pb2
from ...events.v1 import events_pb2 as vega_dot_events_dot_v1_dot_events__pb2
from protoc_gen_openapiv2.options import (
    annotations_pb2 as protoc__gen__openapiv2_dot_options_dot_annotations__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bvega/api/v1/corestate.proto\x12\x0bvega.api.v1\x1a\x11vega/assets.proto\x1a\x15vega/governance.proto\x1a\x12vega/markets.proto\x1a\x0fvega/vega.proto\x1a\x1bvega/events/v1/events.proto\x1a.protoc-gen-openapiv2/options/annotations.proto"{\n\x07\x41\x63\x63ount\x12\x14\n\x05party\x18\x01 \x01(\tR\x05party\x12\x16\n\x06market\x18\x02 \x01(\tR\x06market\x12\x18\n\x07\x62\x61lance\x18\x03 \x01(\tR\x07\x62\x61lance\x12\x14\n\x05\x61sset\x18\x05 \x01(\tR\x05\x61sset\x12\x12\n\x04type\x18\x06 \x01(\tR\x04type"C\n\x13ListAccountsRequest\x12\x14\n\x05party\x18\x01 \x01(\tR\x05party\x12\x16\n\x06market\x18\x02 \x01(\tR\x06market"H\n\x14ListAccountsResponse\x12\x30\n\x08\x61\x63\x63ounts\x18\x01 \x03(\x0b\x32\x14.vega.api.v1.AccountR\x08\x61\x63\x63ounts")\n\x11ListAssetsRequest\x12\x14\n\x05\x61sset\x18\x01 \x01(\tR\x05\x61sset"9\n\x12ListAssetsResponse\x12#\n\x06\x61ssets\x18\x01 \x03(\x0b\x32\x0b.vega.AssetR\x06\x61ssets"R\n\x1cListNetworkParametersRequest\x12\x32\n\x15network_parameter_key\x18\x01 \x01(\tR\x13networkParameterKey"f\n\x1dListNetworkParametersResponse\x12\x45\n\x12network_parameters\x18\x01 \x03(\x0b\x32\x16.vega.NetworkParameterR\x11networkParameters"\x1a\n\x18ListNetworkLimitsRequest"W\n\x19ListNetworkLimitsResponse\x12:\n\x0enetwork_limits\x18\x01 \x01(\x0b\x32\x13.vega.NetworkLimitsR\rnetworkLimits"\x14\n\x12ListPartiesRequest"<\n\x13ListPartiesResponse\x12%\n\x07parties\x18\x01 \x03(\x0b\x32\x0b.vega.PartyR\x07parties"\x17\n\x15ListValidatorsRequest"Y\n\x16ListValidatorsResponse\x12?\n\nvalidators\x18\x01 \x03(\x0b\x32\x1f.vega.events.v1.ValidatorUpdateR\nvalidators",\n\x12ListMarketsRequest\x12\x16\n\x06market\x18\x01 \x01(\tR\x06market"=\n\x13ListMarketsResponse\x12&\n\x07markets\x18\x01 \x03(\x0b\x32\x0c.vega.MarketR\x07markets"N\n\x14ListProposalsRequest\x12\x1a\n\x08proposal\x18\x01 \x01(\tR\x08proposal\x12\x1a\n\x08proposer\x18\x02 \x01(\tR\x08proposer"E\n\x15ListProposalsResponse\x12,\n\tproposals\x18\x01 \x03(\x0b\x32\x0e.vega.ProposalR\tproposals"0\n\x16ListMarketsDataRequest\x12\x16\n\x06market\x18\x01 \x01(\tR\x06market"N\n\x17ListMarketsDataResponse\x12\x33\n\x0cmarkets_data\x18\x01 \x03(\x0b\x32\x10.vega.MarketDataR\x0bmarketsData"D\n\x10ListVotesRequest\x12\x1a\n\x08proposal\x18\x01 \x01(\tR\x08proposal\x12\x14\n\x05party\x18\x02 \x01(\tR\x05party"5\n\x11ListVotesResponse\x12 \n\x05votes\x18\x01 \x03(\x0b\x32\n.vega.VoteR\x05votes"\x9f\x01\n\nPartyStake\x12\x14\n\x05party\x18\x01 \x01(\tR\x05party\x12\x36\n\x17\x63urrent_stake_available\x18\x02 \x01(\tR\x15\x63urrentStakeAvailable\x12\x43\n\x0estake_linkings\x18\x03 \x03(\x0b\x32\x1c.vega.events.v1.StakeLinkingR\rstakeLinkings"/\n\x17ListPartiesStakeRequest\x12\x14\n\x05party\x18\x01 \x01(\tR\x05party"X\n\x18ListPartiesStakeResponse\x12<\n\rparties_stake\x18\x01 \x03(\x0b\x32\x17.vega.api.v1.PartyStakeR\x0cpartiesStake"_\n\x16ListDelegationsRequest\x12\x14\n\x05party\x18\x01 \x01(\tR\x05party\x12\x12\n\x04node\x18\x02 \x01(\tR\x04node\x12\x1b\n\tepoch_seq\x18\x03 \x01(\tR\x08\x65pochSeq"M\n\x17ListDelegationsResponse\x12\x32\n\x0b\x64\x65legations\x18\x01 \x03(\x0b\x32\x10.vega.DelegationR\x0b\x64\x65legations2\xca\x08\n\x10\x43oreStateService\x12S\n\x0cListAccounts\x12 .vega.api.v1.ListAccountsRequest\x1a!.vega.api.v1.ListAccountsResponse\x12M\n\nListAssets\x12\x1e.vega.api.v1.ListAssetsRequest\x1a\x1f.vega.api.v1.ListAssetsResponse\x12n\n\x15ListNetworkParameters\x12).vega.api.v1.ListNetworkParametersRequest\x1a*.vega.api.v1.ListNetworkParametersResponse\x12\x62\n\x11ListNetworkLimits\x12%.vega.api.v1.ListNetworkLimitsRequest\x1a&.vega.api.v1.ListNetworkLimitsResponse\x12P\n\x0bListParties\x12\x1f.vega.api.v1.ListPartiesRequest\x1a .vega.api.v1.ListPartiesResponse\x12Y\n\x0eListValidators\x12".vega.api.v1.ListValidatorsRequest\x1a#.vega.api.v1.ListValidatorsResponse\x12P\n\x0bListMarkets\x12\x1f.vega.api.v1.ListMarketsRequest\x1a .vega.api.v1.ListMarketsResponse\x12V\n\rListProposals\x12!.vega.api.v1.ListProposalsRequest\x1a".vega.api.v1.ListProposalsResponse\x12\\\n\x0fListMarketsData\x12#.vega.api.v1.ListMarketsDataRequest\x1a$.vega.api.v1.ListMarketsDataResponse\x12J\n\tListVotes\x12\x1d.vega.api.v1.ListVotesRequest\x1a\x1e.vega.api.v1.ListVotesResponse\x12_\n\x10ListPartiesStake\x12$.vega.api.v1.ListPartiesStakeRequest\x1a%.vega.api.v1.ListPartiesStakeResponse\x12\\\n\x0fListDelegations\x12#.vega.api.v1.ListDelegationsRequest\x1a$.vega.api.v1.ListDelegationsResponseBoZ,code.vegaprotocol.io/vega/protos/vega/api/v1\x92\x41>\x12#\n\x14Vega core state APIs2\x0bv0.63.0+dev\x1a\x13lb.testnet.vega.xyz*\x02\x01\x02\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "vega.api.v1.corestate_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z,code.vegaprotocol.io/vega/protos/vega/api/v1\222A>\022#\n\024Vega core state APIs2\013v0.63.0+dev\032\023lb.testnet.vega.xyz*\002\001\002"
    _ACCOUNT._serialized_start = 200
    _ACCOUNT._serialized_end = 323
    _LISTACCOUNTSREQUEST._serialized_start = 325
    _LISTACCOUNTSREQUEST._serialized_end = 392
    _LISTACCOUNTSRESPONSE._serialized_start = 394
    _LISTACCOUNTSRESPONSE._serialized_end = 466
    _LISTASSETSREQUEST._serialized_start = 468
    _LISTASSETSREQUEST._serialized_end = 509
    _LISTASSETSRESPONSE._serialized_start = 511
    _LISTASSETSRESPONSE._serialized_end = 568
    _LISTNETWORKPARAMETERSREQUEST._serialized_start = 570
    _LISTNETWORKPARAMETERSREQUEST._serialized_end = 652
    _LISTNETWORKPARAMETERSRESPONSE._serialized_start = 654
    _LISTNETWORKPARAMETERSRESPONSE._serialized_end = 756
    _LISTNETWORKLIMITSREQUEST._serialized_start = 758
    _LISTNETWORKLIMITSREQUEST._serialized_end = 784
    _LISTNETWORKLIMITSRESPONSE._serialized_start = 786
    _LISTNETWORKLIMITSRESPONSE._serialized_end = 873
    _LISTPARTIESREQUEST._serialized_start = 875
    _LISTPARTIESREQUEST._serialized_end = 895
    _LISTPARTIESRESPONSE._serialized_start = 897
    _LISTPARTIESRESPONSE._serialized_end = 957
    _LISTVALIDATORSREQUEST._serialized_start = 959
    _LISTVALIDATORSREQUEST._serialized_end = 982
    _LISTVALIDATORSRESPONSE._serialized_start = 984
    _LISTVALIDATORSRESPONSE._serialized_end = 1073
    _LISTMARKETSREQUEST._serialized_start = 1075
    _LISTMARKETSREQUEST._serialized_end = 1119
    _LISTMARKETSRESPONSE._serialized_start = 1121
    _LISTMARKETSRESPONSE._serialized_end = 1182
    _LISTPROPOSALSREQUEST._serialized_start = 1184
    _LISTPROPOSALSREQUEST._serialized_end = 1262
    _LISTPROPOSALSRESPONSE._serialized_start = 1264
    _LISTPROPOSALSRESPONSE._serialized_end = 1333
    _LISTMARKETSDATAREQUEST._serialized_start = 1335
    _LISTMARKETSDATAREQUEST._serialized_end = 1383
    _LISTMARKETSDATARESPONSE._serialized_start = 1385
    _LISTMARKETSDATARESPONSE._serialized_end = 1463
    _LISTVOTESREQUEST._serialized_start = 1465
    _LISTVOTESREQUEST._serialized_end = 1533
    _LISTVOTESRESPONSE._serialized_start = 1535
    _LISTVOTESRESPONSE._serialized_end = 1588
    _PARTYSTAKE._serialized_start = 1591
    _PARTYSTAKE._serialized_end = 1750
    _LISTPARTIESSTAKEREQUEST._serialized_start = 1752
    _LISTPARTIESSTAKEREQUEST._serialized_end = 1799
    _LISTPARTIESSTAKERESPONSE._serialized_start = 1801
    _LISTPARTIESSTAKERESPONSE._serialized_end = 1889
    _LISTDELEGATIONSREQUEST._serialized_start = 1891
    _LISTDELEGATIONSREQUEST._serialized_end = 1986
    _LISTDELEGATIONSRESPONSE._serialized_start = 1988
    _LISTDELEGATIONSRESPONSE._serialized_end = 2065
    _CORESTATESERVICE._serialized_start = 2068
    _CORESTATESERVICE._serialized_end = 3166
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/markets.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import data_source_pb2 as vega_dot_data__source__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12vega/markets.proto\x12\x04vega\x1a\x16vega/data_source.proto"E\n\x0f\x41uctionDuration\x12\x1a\n\x08\x64uration\x18\x01 \x01(\x03R\x08\x64uration\x12\x16\n\x06volume\x18\x02 \x01(\x04R\x06volume"\xbc\x03\n\x06\x46uture\x12)\n\x10settlement_asset\x18\x01 \x01(\tR\x0fsettlementAsset\x12\x1d\n\nquote_name\x18\x02 \x01(\tR\tquoteName\x12\x63\n$data_source_spec_for_settlement_data\x18\x03 \x01(\x0b\x32\x14.vega.DataSourceSpecR\x1f\x64\x61taSourceSpecForSettlementData\x12k\n(data_source_spec_for_trading_termination\x18\x04 \x01(\x0b\x32\x14.vega.DataSourceSpecR#dataSourceSpecForTradingTermination\x12\\\n\x18\x64\x61ta_source_spec_binding\x18\x05 \x01(\x0b\x32#.vega.DataSourceSpecToFutureBindingR\x15\x64\x61taSourceSpecBinding\x12\x38\n\x18settlement_data_decimals\x18\x06 \x01(\rR\x16settlementDataDecimals"\x9b\x01\n\x1d\x44\x61taSourceSpecToFutureBinding\x12\x38\n\x18settlement_data_property\x18\x01 \x01(\tR\x16settlementDataProperty\x12@\n\x1ctrading_termination_property\x18\x02 \x01(\tR\x1atradingTerminationProperty"(\n\x12InstrumentMetadata\x12\x12\n\x04tags\x18\x01 \x03(\tR\x04tags"\xad\x01\n\nInstrument\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04\x63ode\x18\x02 \x01(\tR\x04\x63ode\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\x12\x34\n\x08metadata\x18\x04 \x01(\x0b\x32\x18.vega.InstrumentMetadataR\x08metadata\x12&\n\x06\x66uture\x18\x64 \x01(\x0b\x32\x0c.vega.FutureH\x00R\x06\x66utureB\t\n\x07product"\x92\x01\n\x12LogNormalRiskModel\x12\x36\n\x17risk_aversion_parameter\x18\x01 \x01(\x01R\x15riskAversionParameter\x12\x10\n\x03tau\x18\x02 \x01(\x01R\x03tau\x12\x32\n\x06params\x18\x03 \x01(\x0b\x32\x1a.vega.LogNormalModelParamsR\x06params"J\n\x14LogNormalModelParams\x12\x0e\n\x02mu\x18\x01 \x01(\x01R\x02mu\x12\x0c\n\x01r\x18\x02 \x01(\x01R\x01r\x12\x14\n\x05sigma\x18\x03 \x01(\x01R\x05sigma"B\n\x0fSimpleRiskModel\x12/\n\x06params\x18\x01 \x01(\x0b\x32\x17.vega.SimpleModelParamsR\x06params"\xd1\x01\n\x11SimpleModelParams\x12\x1f\n\x0b\x66\x61\x63tor_long\x18\x01 \x01(\x01R\nfactorLong\x12!\n\x0c\x66\x61\x63tor_short\x18\x02 \x01(\x01R\x0b\x66\x61\x63torShort\x12\x1e\n\x0bmax_move_up\x18\x03 \x01(\x01R\tmaxMoveUp\x12"\n\rmin_move_down\x18\x04 \x01(\x01R\x0bminMoveDown\x12\x34\n\x16probability_of_trading\x18\x05 \x01(\x01R\x14probabilityOfTrading"\x89\x01\n\x0eScalingFactors\x12!\n\x0csearch_level\x18\x01 \x01(\x01R\x0bsearchLevel\x12%\n\x0einitial_margin\x18\x02 \x01(\x01R\rinitialMargin\x12-\n\x12\x63ollateral_release\x18\x03 \x01(\x01R\x11\x63ollateralRelease"Q\n\x10MarginCalculator\x12=\n\x0fscaling_factors\x18\x01 \x01(\x0b\x32\x14.vega.ScalingFactorsR\x0escalingFactors"\xad\x02\n\x12TradableInstrument\x12\x30\n\ninstrument\x18\x01 \x01(\x0b\x32\x10.vega.InstrumentR\ninstrument\x12\x43\n\x11margin_calculator\x18\x02 \x01(\x0b\x32\x16.vega.MarginCalculatorR\x10marginCalculator\x12M\n\x15log_normal_risk_model\x18\x64 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\x12logNormalRiskModel\x12\x43\n\x11simple_risk_model\x18\x65 \x01(\x0b\x32\x15.vega.SimpleRiskModelH\x00R\x0fsimpleRiskModelB\x0c\n\nrisk_model"}\n\nFeeFactors\x12\x1b\n\tmaker_fee\x18\x01 \x01(\tR\x08makerFee\x12-\n\x12infrastructure_fee\x18\x02 \x01(\tR\x11infrastructureFee\x12#\n\rliquidity_fee\x18\x03 \x01(\tR\x0cliquidityFee"2\n\x04\x46\x65\x65s\x12*\n\x07\x66\x61\x63tors\x18\x01 \x01(\x0b\x32\x10.vega.FeeFactorsR\x07\x66\x61\x63tors"\x81\x01\n\x16PriceMonitoringTrigger\x12\x18\n\x07horizon\x18\x01 \x01(\x03R\x07horizon\x12 \n\x0bprobability\x18\x02 \x01(\tR\x0bprobability\x12+\n\x11\x61uction_extension\x18\x03 \x01(\x03R\x10\x61uctionExtension"U\n\x19PriceMonitoringParameters\x12\x38\n\x08triggers\x18\x01 \x03(\x0b\x32\x1c.vega.PriceMonitoringTriggerR\x08triggers"Z\n\x17PriceMonitoringSettings\x12?\n\nparameters\x18\x01 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\nparameters"\xcc\x01\n\x1dLiquidityMonitoringParameters\x12S\n\x17target_stake_parameters\x18\x01 \x01(\x0b\x32\x1b.vega.TargetStakeParametersR\x15targetStakeParameters\x12)\n\x10triggering_ratio\x18\x02 \x01(\x01R\x0ftriggeringRatio\x12+\n\x11\x61uction_extension\x18\x03 \x01(\x03R\x10\x61uctionExtension"_\n\x15TargetStakeParameters\x12\x1f\n\x0btime_window\x18\x01 \x01(\x03R\ntimeWindow\x12%\n\x0escaling_factor\x18\x02 \x01(\x01R\rscalingFactor"\xc0\x08\n\x06Market\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12I\n\x13tradable_instrument\x18\x02 \x01(\x0b\x32\x18.vega.TradableInstrumentR\x12tradableInstrument\x12%\n\x0e\x64\x65\x63imal_places\x18\x03 \x01(\x04R\rdecimalPlaces\x12\x1e\n\x04\x66\x65\x65s\x18\x04 \x01(\x0b\x32\n.vega.FeesR\x04\x66\x65\x65s\x12>\n\x0fopening_auction\x18\x05 \x01(\x0b\x32\x15.vega.AuctionDurationR\x0eopeningAuction\x12Y\n\x19price_monitoring_settings\x18\x06 \x01(\x0b\x32\x1d.vega.PriceMonitoringSettingsR\x17priceMonitoringSettings\x12k\n\x1fliquidity_monitoring_parameters\x18\x07 \x01(\x0b\x32#.vega.LiquidityMonitoringParametersR\x1dliquidityMonitoringParameters\x12;\n\x0ctrading_mode\x18\x08 \x01(\x0e\x32\x18.vega.Market.TradingModeR\x0btradingMode\x12(\n\x05state\x18\t \x01(\x0e\x32\x12.vega.Market.StateR\x05state\x12\x43\n\x11market_timestamps\x18\n \x01(\x0b\x32\x16.vega.MarketTimestampsR\x10marketTimestamps\x12\x36\n\x17position_decimal_places\x18\x0b \x01(\x03R\x15positionDecimalPlaces"\xd8\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x12\n\x0eSTATE_PROPOSED\x10\x01\x12\x12\n\x0eSTATE_REJECTED\x10\x02\x12\x11\n\rSTATE_PENDING\x10\x03\x12\x13\n\x0fSTATE_CANCELLED\x10\x04\x12\x10\n\x0cSTATE_ACTIVE\x10\x05\x12\x13\n\x0fSTATE_SUSPENDED\x10\x06\x12\x10\n\x0cSTATE_CLOSED\x10\x07\x12\x1c\n\x18STATE_TRADING_TERMINATED\x10\x08\x12\x11\n\rSTATE_SETTLED\x10\t"\xcc\x01\n\x0bTradingMode\x12\x1c\n\x18TRADING_MODE_UNSPECIFIED\x10\x00\x12\x1b\n\x17TRADING_MODE_CONTINUOUS\x10\x01\x12\x1e\n\x1aTRADING_MODE_BATCH_AUCTION\x10\x02\x12 \n\x1cTRADING_MODE_OPENING_AUCTION\x10\x03\x12#\n\x1fTRADING_MODE_MONITORING_AUCTION\x10\x04\x12\x1b\n\x17TRADING_MODE_NO_TRADING\x10\x05"r\n\x10MarketTimestamps\x12\x1a\n\x08proposed\x18\x01 \x01(\x03R\x08proposed\x12\x18\n\x07pending\x18\x02 \x01(\x03R\x07pending\x12\x12\n\x04open\x18\x03 \x01(\x03R\x04open\x12\x14\n\x05\x63lose\x18\x04 \x01(\x03R\x05\x63loseB\'Z%code.vegaprotocol.io/vega/protos/vegab\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "vega.markets_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z%code.vegaprotocol.io/vega/protos/vega"
    _AUCTIONDURATION._serialized_start = 52
    _AUCTIONDURATION._serialized_end = 121
    _FUTURE._serialized_start = 124
    _FUTURE._serialized_end = 568
    _DATASOURCESPECTOFUTUREBINDING._serialized_start = 571
    _DATASOURCESPECTOFUTUREBINDING._serialized_end = 726
    _INSTRUMENTMETADATA._serialized_start = 728
    _INSTRUMENTMETADATA._serialized_end = 768
    _INSTRUMENT._serialized_start = 771
    _INSTRUMENT._serialized_end = 944
    _LOGNORMALRISKMODEL._serialized_start = 947
    _LOGNORMALRISKMODEL._serialized_end = 1093
    _LOGNORMALMODELPARAMS._serialized_start = 1095
    _LOGNORMALMODELPARAMS._serialized_end = 1169
    _SIMPLERISKMODEL._serialized_start = 1171
    _SIMPLERISKMODEL._serialized_end = 1237
    _SIMPLEMODELPARAMS._serialized_start = 1240
    _SIMPLEMODELPARAMS._serialized_end = 1449
    _SCALINGFACTORS._serialized_start = 1452
    _SCALINGFACTORS._serialized_end = 1589
    _MARGINCALCULATOR._serialized_start = 1591
    _MARGINCALCULATOR._serialized_end = 1672
    _TRADABLEINSTRUMENT._serialized_start = 1675
    _TRADABLEINSTRUMENT._serialized_end = 1976
    _FEEFACTORS._serialized_start = 1978
    _FEEFACTORS._serialized_end = 2103
    _FEES._serialized_start = 2105
    _FEES._serialized_end = 2155
    _PRICEMONITORINGTRIGGER._serialized_start = 2158
    _PRICEMONITORINGTRIGGER._serialized_end = 2287
    _PRICEMONITORINGPARAMETERS._serialized_start = 2289
    _PRICEMONITORINGPARAMETERS._serialized_end = 2374
    _PRICEMONITORINGSETTINGS._serialized_start = 2376
    _PRICEMONITORINGSETTINGS._serialized_end = 2466
    _LIQUIDITYMONITORINGPARAMETERS._serialized_start = 2469
    _LIQUIDITYMONITORINGPARAMETERS._serialized_end = 2673
    _TARGETSTAKEPARAMETERS._serialized_start = 2675
    _TARGETSTAKEPARAMETERS._serialized_end = 2770
    _MARKET._serialized_start = 2773
    _MARKET._serialized_end = 3861
    _MARKET_STATE._serialized_start = 3438
    _MARKET_STATE._serialized_end = 3654
    _MARKET_TRADINGMODE._serialized_start = 3657
    _MARKET_TRADINGMODE._serialized_end = 3861
    _MARKETTIMESTAMPS._serialized_start = 3863
    _MARKETTIMESTAMPS._serialized_end = 3977
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from collections import namedtuple
from typing import Callable, Dict, Iterable, List, Optional, TypeVar, Union

import vega_sim.grpc.client as vac
import vega_sim.proto.data_node.api.v2 as data_node_protos_v2
import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.events.v1.events_pb2 as events_protos

MarketAccount = namedtuple("MarketAccount", ["insurance", "liquidity_fee"])

T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U")


def unroll_v2_pagination(
    base_request: S,
    request_func: Callable[[S], T],
    extraction_func: Callable[[S], List[U]],
) -> List[T]:
    base_request.pagination.CopyFrom(data_node_protos_v2.trading_data.Pagination())
    response = request_func(base_request)
    full_list = extraction_func(response)
    while response.page_info.has_next_page:
        base_request.pagination.after = response.page_info.end_cursor
        response = request_func(base_request)
        curr_list = extraction_func(response)
        full_list.extend(curr_list)
    return full_list


def positions_by_market(
    pub_key: str,
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.vega.Position]:
    """Output positions of a party."""
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListPositionsRequest(
            party_id=pub_key,
            market_id=market_id,
        ),
        request_func=lambda x: data_client.ListPositions(x).positions,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def all_markets(
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.markets.Market]:
    """
    Output market info.
    """
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListMarketsRequest(),
        request_func=lambda x: data_client.ListMarkets(x).markets,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def market_info(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vega_protos.markets.Market:
    """
    Output market info.
    """
    return data_client.GetMarket(
        data_node_protos_v2.trading_data.GetMarketRequest(market_id=market_id)
    ).market


def list_assets(data_client: vac.VegaTradingDataClientV2):
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAssetsRequest(),
        request_func=lambda x: data_client.ListAssets(x).assets,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def asset_info(
    asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vac.vega.assets.Asset:
    """Returns information on a given asset selected by its ID

    Args:
        asset_id:
            str, The ID of the asset requested
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC data client

    Returns:
        vega.assets.Asset, Object containing data about the requested asset
    """
    return data_client.GetAsset(
        data_node_protos_v2.trading_data.GetAssetRequest(asset_id=asset_id)
    ).asset


def all_market_accounts(
    asset_id: str,
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> Dict[vega_protos.vega.AccountType, MarketAccount]:
    """
    Output liquidity fee account/ insurance pool in the market
    """

    account_filter = data_node_protos_v2.trading_data.AccountFilter(
        asset_id=asset_id,
        market_ids=[market_id],
    )

    accounts = unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAccountsRequest(
            filter=account_filter
        ),
        request_func=lambda x: data_client.ListAccounts(x).accounts,
        extraction_func=lambda res: [i.account for i in res.edges],
    )

    return {account.type: account for account in accounts}


def market_accounts(
    asset_id: str,
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> MarketAccount:
    """
    Output liquidity fee account/ insurance pool in the market
    """
    account_type_map = all_market_accounts(
        asset_id=asset_id, market_id=market_id, data_client=data_client
    )
    return MarketAccount(
        account_type_map[vega_protos.vega.ACCOUNT_TYPE_INSURANCE],
        account_type_map[vega_protos.vega.ACCOUNT_TYPE_FEES_LIQUIDITY],
    )


def party_accounts(
    asset_id: str,
    party_id: str,
    data_client: vac.VegaTradingDataClientV2,
):

    account_filter = data_node_protos_v2.trading_data.AccountFilter(
        asset_id=asset_id,
        party_ids=[party_id],
    )

    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAccountsRequest(
            filter=account_filter
        ),
        request_func=lambda x: data_client.ListAccounts(x).accounts,
        extraction_func=lambda res: [i.account for i in res.edges],
    )


def market_data(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vega_protos.vega.MarketData:
    """
    Output market info.
    """
    return data_client.GetLatestMarketData(
        data_node_protos_v2.trading_data.GetLatestMarketDataRequest(market_id=market_id)
    ).market_data


def infrastructure_fee_accounts(
    asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.vega.Account]:
    """
    Output infrastructure fee accounts
    """

    account_filter = data_node_protos_v2.trading_data.AccountFilter(
        asset_id=asset_id,
        account_types=[
            vega_protos.vega.AccountType.ACCOUNT_TYPE_FEES_INFRASTRUCTURE,
        ],
    )

    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAccountsRequest(
            filter=account_filter
        ),
        request_func=lambda x: data_client.ListAccounts(x).accounts,
        extraction_func=lambda res: [i.account for i in res.edges],
    )


def list_orders(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str = None,
    party_id: str = None,
    reference: str = None,
    live_only: bool = True,
) -> List[vega_protos.vega.Orders]:
    """Gets a list of Orders for the specified market and party.

    Function queries the datanode for a response of orders for the specified market_id
    and party_id and state (defaults to live orders only). The function then unrolls
    pagination and returns a list of Orders.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            An instantiated gRPC trading_data_client_V2.
        market_id (str):
            Id for market to return orders from.
        party_id (str):
            Id for party to return orders from.
        live_only (bool, optional):
            Whether to only return live orders. Defaults to True.

    Returns:
        _type_: _description_
    """

    request = data_node_protos_v2.trading_data.ListOrdersRequest(live_only=live_only)

    for attr, val in [
        ("market_id", market_id),
        ("party_id", party_id),
        ("reference", reference),
    ]:
        if val is not None:
            setattr(request, attr, val)

    return unroll_v2_pagination(
        base_request=request,
        request_func=lambda x: data_client.ListOrders(x).orders,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def order_status(
    order_id: str, data_client: vac.VegaTradingDataClientV2, version: int = 0
) -> Optional[vega_protos.vega.Order]:
    """Loads information about a specific order identified by the ID.
    Optionally return historic order versions.

    Args:
        order_id:
            str, the order identifier as specified by Vega when originally placed
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC trading data client
        version:
            int, Optional, Version of the order:
                - Set `version` to 0 for most recent version of the order
                - Set `1` for original version of the order
                - Set `2` for first amendment, `3` for second amendment, etc
    Returns:
        Optional[vega.Order], the requested Order object or None if nothing found
    """
    return data_client.GetOrder(
        data_node_protos_v2.trading_data.GetOrderRequest(
            order_id=order_id, version=version
        )
    ).order


def market_depth(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    max_depth: Optional[int] = None,
) -> Optional[vega_protos.vega.MarketDepth]:
    return data_client.GetLatestMarketDepth(
        data_node_protos_v2.trading_data.GetLatestMarketDepthRequest(
            market_id=market_id, max_depth=max_depth
        )
    )


def liquidity_provisions(
    data_client: vac.VegaTradingDataClientV2,
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
) -> Optional[List[vega_protos.vega.LiquidityProvision]]:
    """Loads the current liquidity provision(s) for a given market and/or party.

    Args:
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC trading data client
        market_id:
            Optional[str], the ID of the market from which to pull liquidity provisions
        party_id:
            Optional[str], the ID of the party from which to pull liquidity provisions

    Returns:
        List[LiquidityProvision], list of liquidity provisions (if any exist)
    """
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListLiquidityProvisionsRequest(
            market_id=market_id,
            party_id=party_id,
        ),
        request_func=lambda x: data_client.ListLiquidityProvisions(
            x
        ).liquidity_provisions,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def order_subscription(
    data_client: vac.VegaCoreClient,
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
) -> Iterable[vega_protos.api.v1.core.ObserveEventBusResponse]:
    """Subscribe to a stream of Order updates from the data-node.
    The stream of orders returned from this function is an iterable which
    does not end and will continue to tick another order update whenever
    one is received.

    Args:
        market_id:
            Optional[str], If provided, only update orders from this market
        party_id:
            Optional[str], If provided, only update orders from this party
    Returns:
        Iterable[List[vega.Order]], Infinite iterable of lists of order updates
    """
    return data_client.ObserveEventBus(
        iter(
            [
                vega_protos.api.v1.core.ObserveEventBusRequest(
                    type=[events_protos.BUS_EVENT_TYPE_ORDER]
                )
            ]
        )
    )


def margin_levels(
    data_client: vac.VegaTradingDataClientV2,
    party_id: str,
    market_id: Optional[str] = None,
) -> List[vega_protos.vega.MarginLevels]:
    return unroll_v2_pagination(
        data_node_protos_v2.trading_data.ListMarginLevelsRequest(
            market_id=market_id, party_id=party_id
        ),
        request_func=lambda x: data_client.ListMarginLevels(x).margin_levels,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def get_trades(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    party_id: Optional[str] = None,
    order_id: Optional[str] = None,
) -> List[vega_protos.vega.Trade]:
    return unroll_v2_pagination(
        data_node_protos_v2.trading_data.ListTradesRequest(
            market_id=market_id, party_id=party_id, order_id=order_id
        ),
        request_func=lambda x: data_client.ListTrades(x).trades,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def get_network_parameter(
    data_client: vac.VegaTradingDataClientV2,
    key: str,
) -> Union[str, int, float]:
    """Returns the value of the specified network parameter.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            Instantiated gRPC client
        key (str):
            The key identifying the network parameter.

    Returns:
        Any:
            The value of the specified network parameter.
    """
    return data_client.GetNetworkParameter(
        data_node_protos_v2.trading_data.GetNetworkParameterRequest(key=key),
    ).network_parameter



import abc
import numpy as np
from typing import Optional

from vega_sim.null_service import VegaServiceNull
from vega_sim.scenario.constants import Network


class Scenario(abc.ABC):
    def run_iteration(
        self,
        vega: VegaServiceNull,
        network: Optional[Network],
        pause_at_completion: bool = False,
        random_state: Optional[np.random.RandomState] = None,
    ):
        pass



"""network_service.py

This module contains functions and classes for the creation of the
VegaServiceNetwork class. A VegaServiceNetwork object can be used to
communicate with either a public or internal Vega network and run a
Vega Wallet service for communication with the chosen Vega network.
The VegaServiceNetwork inherits properties and methods from the VegaService
class. Inherited methods can be used to communicate with the Vega datanode
and Vega wallet services. Redundant properties and methods are overwritten.
Example:

    For an example, try running the below command. It will create a
    VegaServiceNetwork object, the relevant services and try sending some
    commands.

        $ python -m vega_sim.network_service

    If the above example is throwing errors, check you have cloned the
    required bin folders with the following command.

        $ make networks

Todo:
    • Test all public and internal Vega networks are compatible.
    • Test all market sim agents are compatible.
"""

import grpc
import toml
import time
import signal
import logging
import tempfile
import itertools
import subprocess
import webbrowser
import multiprocessing

from typing import Optional

from os import getcwd, path, environ

import vega_sim.grpc.client as vac

from vega_sim import vega_bin_path
from vega_sim.service import VegaService
from vega_sim.wallet.base import Wallet
from vega_sim.wallet.vega_wallet import VegaWallet
from vega_sim.null_service import find_free_port, _popen_process

from vega_sim.constants import DATA_NODE_GRPC_PORT
from vega_sim.scenario.constants import Network

logger = logging.getLogger(__name__)


def add_network_config(
    network_config_path: str,
):
    """Creates a network config file for the specified network.
    Method creates a copy of the config file specified in the local vega path
    networks folder. This config file is used to run a wallet service.

    Args:
        network_config_path (str):
            Path to the network config file to create a local network from.

    Raises:
        ValueError:
            If no file exists at the specified path.
    """

    if not path.exists(network_config_path):
        raise ValueError(f"No network config file at the specified path.")

    args = [
        "vega",
        "wallet",
        "network",
        "import",
        "--from-file",
        network_config_path,
        "--force",
    ]

    subprocess.call(args=args)


def manage_vega_processes(
    network: str,
    vega_console_path: str,
    vega_console_port: int,
    data_node_query_url: str,
    wallet_url: str,
    run_with_wallet: Optional[bool] = True,
    run_with_console: Optional[bool] = True,
    log_dir: Optional[str] = None,
):

    processes = []

    tmp_vega_dir = tempfile.mkdtemp(prefix="vega-sim-") if log_dir is None else log_dir

    if run_with_wallet:
        vegaWalletProcess = _popen_process(
            popen_args=[
                "vega",
                "wallet",
                "service",
                "run",
                "--network",
                network,
                "--automatic-consent",
                "--no-version-check",
            ],
            dir_root=tmp_vega_dir,
            log_name="vegawallet",
        )
        processes.append(vegaWalletProcess)

    if run_with_console:
        env_copy = environ.copy()
        env_copy.update(
            {
                "NX_VEGA_URL": data_node_query_url,
                "NX_VEGA_WALLET_URL": wallet_url,
                "NX_VEGA_ENV": "CUSTOM",
                "NX_PORT": f"{vega_console_port}",
            }
        )

        vegaConsoleProcess = _popen_process(
            popen_args=[
                "yarn",
                "--cwd",
                vega_console_path,
                "nx",
                "serve",
                "--port",
                f"{vega_console_port}",
                "-o",
                "trading",
            ],
            dir_root=tmp_vega_dir,
            log_name="console",
            env=env_copy,
        )
        processes.append(vegaConsoleProcess)

    signal.sigwait([signal.SIGKILL, signal.SIGTERM])
    for process in processes:
        process.terminate()
    for process in processes:
        return_code = process.poll()
        if return_code is not None:
            continue
        time.sleep(5)
        process.kill()


class VegaServiceNetwork(VegaService):
    """Class for handling services for communicating with a Vega network."""

    def __init__(
        self,
        network: Network,
        run_with_wallet: bool = True,
        run_with_console: bool = True,
    ):
        """Method initialises the class.

        Args:
            network (Network):
                Defines which network to connect service to.
            run_with_wallet (bool, optional):
                Defines whether to start a wallet process.
            run_with_console (bool, optional):
                Defines whether to start a console process.
        """

        # Run init method inherited from VegaService with network arguments.
        super().__init__(can_control_time=False, warn_on_raw_data_access=False)

        self.network = network
        self.run_with_wallet = run_with_wallet
        self.run_with_console = run_with_console

        self._wallet = None
        self._wallet_url = None
        self._data_node_grpc_url = None
        self._data_node_query_url = None
        self._network_config = None

        self.vega_console_path = path.join(vega_bin_path, "console")

        self.log_dir = tempfile.mkdtemp(prefix="vega-sim-")

    def __enter__(self):
        """Defines behaviour when class entered by a with statement."""
        self.start()

        return self

    def __exit__(self, type, value, traceback):
        """Defines behaviour when class exited by a with statement."""
        self.stop()

    def start(self):

        ctx = multiprocessing.get_context()
        vega_console_port = find_free_port()
        self.proc = ctx.Process(
            target=manage_vega_processes,
            kwargs={
                "network": self.network.value,
                "log_dir": self.log_dir,
                "run_with_wallet": self.run_with_wallet,
                "run_with_console": self.run_with_console,
                "vega_console_path": self.vega_console_path,
                "vega_console_port": vega_console_port,
                "data_node_query_url": self.data_node_query_url,
                "wallet_url": self.wallet_url,
            },
        )
        self.proc.start()

        if self.run_with_console:
            logger.info(
                "Vega Running. Console launched at"
                f" http://localhost:{vega_console_port}"
            )
            webbrowser.open(f"http://localhost:{vega_console_port}/", new=2)

    def stop(self) -> None:

        super().stop()
        if self.proc is None:
            logger.info("Stop called but nothing to stop")
        else:
            self.proc.terminate()

    def wait_fn(self, wait_multiple: float = 1) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling time."
        )

    def wait_for_datanode_sync(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling time."
        )

    def wait_for_core_catchup(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling time."
        )

    def wait_for_total_catchup(self) -> None:
        """Overrides redundant parent method."""
        logging.debug(
            "Parent method overridden as VegaNetworkService incapable of controlling time."
        )

    @property
    def network_config(self) -> dict:
        if self._network_config is None:

            public_path = path.join(
                getcwd(),
                "vega_sim",
                "bin",
                "networks-internal",
                self.network.name.lower(),
                f"{self.network.value}.toml",
            )
            internal_path = path.join(
                getcwd(),
                "vega_sim",
                "bin",
                "networks",
                self.network.name.lower(),
                f"{self.network.value}.toml",
            )

            if path.exists(public_path):
                self._network_config = toml.load(public_path)
                add_network_config(public_path)

            elif path.exists(internal_path):
                self._network_config = toml.load(internal_path)
                add_network_config(internal_path)

            else:
                raise ValueError(
                    f"ERROR! {self.network.name.lower()} network does not exist"
                )

        return self._network_config

    @property
    def data_node_grpc_url(self) -> str:
        if self._data_node_grpc_url is None:
            url = self.network_config["API"]["GRPC"]["Hosts"][0]
            self._data_node_grpc_url = f"{url.split(':')[0]}:{DATA_NODE_GRPC_PORT}"
        return self._data_node_grpc_url

    @property
    def data_node_query_url(self) -> str:
        if self._data_node_query_url is None:
            url = self.network_config["API"]["GraphQL"]["Hosts"][0]
            self._data_node_query_url = url
        return self._data_node_query_url

    @property
    def wallet_url(self) -> str:
        if self._wallet_url is None:
            self._wallet_url = (
                f"http://{self.network_config['Host']}:{self.network_config['Port']}"
            )
        return self._wallet_url

    @property
    def wallet(self) -> Wallet:
        if self._wallet is None:
            self._wallet = VegaWallet(self.wallet_url)
        return self._wallet

    @property
    def core_state_client(self) -> None:
        logging.debug(
            "Parent property overridden as VegaNetworkService does not need a core client.",
        )
        pass

    @property
    def core_client(self) -> None:
        logging.debug(
            "Parent property overridden as VegaNetworkService does not need a core client.",
        )
        pass

    def check_datanode(self, raise_on_error: Optional[bool] = True):
        """Checks if the current data-node connection is healthy.

        If the current data-node connection has timed out or the end-point is
        unresponsive then the service will attempt to establish a connection
        with a new healthy data-node.

        Args:
            max_attempts (int, optional):
                Maximum number of connection attempts to attempt before raising an
                error. Defaults to -1 (infinite attempts).
        """

        try:
            self.ping_datanode()
            return

        except grpc.FutureTimeoutError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} timed out."
                )
                self.switch_datanode()

        except grpc._channel._InactiveRpcError as e:
            if raise_on_error:
                raise e
            else:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} inactive."
                )
                self.switch_datanode()

    def switch_datanode(self, max_attempts: Optional[int] = -1):
        """Attempts to establish a new data-node connection.

        Args:
            max_attempts (int, optional):
                Maximum number of connection attempts to attempt before raising an
                error. Defaults to -1 (infinite attempts).
        """

        attempts = 0
        for url in itertools.cycle(self.network_config["API"]["GRPC"]["Hosts"]):
            attempts += 1

            try:
                self._data_node_grpc_url = f"{url.split(':')[0]}:{DATA_NODE_GRPC_PORT}"

                logging.debug(f"Switched to endpoint {self._data_node_grpc_url}")

                channel = grpc.insecure_channel(
                    self.data_node_grpc_url, options=(("grpc.enable_http_proxy", 0),)
                )
                grpc.channel_ready_future(channel).result(timeout=30)
                self._trading_data_client_v2 = vac.VegaTradingDataClientV2(
                    self.data_node_grpc_url,
                    channel=channel,
                )

                return

            except grpc.FutureTimeoutError:
                logging.warning(
                    f"Connection to endpoint {self._data_node_grpc_url} timed out."
                )

            if attempts == max_attempts:
                break

        raise Exception("Unable to establish connection to a data-node.")


if __name__ == "__main__":
    """Module Example"""

    log = logging.basicConfig(level=logging.INFO)

    # Create a service connected to the fairground network.
    with VegaServiceNetwork(
        network=Network.FAIRGROUND,
        run_with_wallet=True,
        run_with_console=True,
    ) as vega:

        # Show all the markets on the network
        markets = vega.all_markets()
        logging.info(markets)

        # Show data for a specific market
        market = vega.market_data(market_id=markets[0].id)

    # Create a service connected to the stagnet3 network.
    with VegaServiceNetwork(
        network=Network.FAIRGROUND,
        run_with_wallet=True,
        run_with_console=True,
    ) as vega:

        # Show all the markets on the network
        markets = vega.all_markets()
        logging.info(markets)

        # Show data for a specific market
        market = vega.market_data(market_id=markets[0].id)



from abc import abstractmethod
from dataclasses import dataclass
import numpy as np
from collections import namedtuple, defaultdict
from typing import List, Tuple

from functools import partial
from tqdm import tqdm

import torch
import torch.nn as nn
from torch.distributions.categorical import Categorical

from vega_sim.reinforcement.networks import (
    Softmax,
    FFN,
    FFN_Params_Normal,
    FFN_Q,
)
from vega_sim.reinforcement.helpers import apply_funcs, to_torch, toggle
from vega_sim.reinforcement.la_market_state import LAMarketState
from vega_sim.reinforcement.la_market_state import AbstractAction
from vega_sim.reinforcement.la_market_state import states_to_sarsa

from vega_sim.reinforcement.distributions import (
    lognorm_sample,
    lognorm_logprob,
    reg_policy,
)

from vega_sim.api.helpers import num_from_padded_int

from vega_sim.environment.agent import Agent
from vega_sim.environment import VegaState
from vega_sim.environment.agent import StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import markets as markets_protos

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Send selling/buying MOs to hit LP orders
WALLET = WalletConfig("learner", "learner")


def state_fn(
    service: VegaServiceNull,
    agents: List[Agent],
    state_values=None,
) -> Tuple[LAMarketState, AbstractAction]:
    learner = [a for a in agents if isinstance(a, LearningAgent)][0]
    return (learner.latest_state, learner.latest_action)


class LearningAgent(StateAgentWithWallet):
    @abstractmethod
    def __init__(
        self,
        device: str,
        logfile_pol_imp: str,
        logfile_pol_eval: str,
        logfile_pnl: str,
        discount_factor: float,
        num_levels: int,
        wallet_name: str,
        wallet_pass: str,
        market_name: str,
        initial_balance: int,
        position_decimals: int,
        inventory_penalty: float = 0.0,
    ):
        super().__init__(wallet_name=wallet_name, wallet_pass=wallet_pass)

        self.step_num = 0
        self.latest_state = None
        self.latest_action = None
        self.device = device
        self.discount_factor = discount_factor
        self.initial_balance = initial_balance

        self.memory = defaultdict(list)
        self.memory_capacity = 100_000

        # Coefficients for regularisation
        self.coefH_discr = 1.0
        self.coefH_cont = 0.01
        # losses logger
        self.losses = defaultdict(list)
        # logfile
        self.logfile_pol_imp = logfile_pol_imp
        self.logfile_pol_eval = logfile_pol_eval
        self.logfile_pnl = logfile_pnl

        self.lerningIteration = 0
        self.market_name = market_name
        self.position_decimals = position_decimals
        self.inventory_penalty = inventory_penalty

    def set_market_tag(self, tag: str):
        self.tag = tag

    @abstractmethod
    def move_to_device(self):
        pass

    @abstractmethod
    def move_to_cpu(self):
        pass

    def initialise(self, vega: VegaServiceNull):
        # Initialise wallet
        super().initialise(vega=vega)
        self.step_num = 0
        market_name = self.market_name + f"_{self.tag}"

        # Get market id
        self.market_id = [
            m.id
            for m in self.vega.all_markets()
            if m.tradable_instrument.instrument.name == market_name
        ][0]
        # Get asset id
        self.tdai_id = self.vega.find_asset_id(symbol=f"tDAI_{self.tag}")
        # Top up asset
        self.vega.mint(
            self.wallet_name,
            asset=self.tdai_id,
            amount=self.initial_balance,
        )
        self.vega.wait_fn(2)

    def _update_memory(
        self,
        state: LAMarketState,
        action: AbstractAction,
        reward: float,
        next_state: LAMarketState,
    ):
        pass

    def update_memory(self, states: List[Tuple[LAMarketState, AbstractAction]]):
        """
        Updates memory of the agent, and removes old tuples (s,a,r,s) if memory exceeds its capacity
        """
        for res in states_to_sarsa(states, inventory_penalty=self.inventory_penalty):
            self._update_memory(res[0], res[1], res[2], res[3])
        # remove old tuples if memory exceeds its capaciy
        for key, value in self.memory.items():
            if len(self.memory[key]) > self.memory_capacity:
                first_n = len(self.memory[key]) - self.memory_capacity
                del self.memory[key][:first_n]
        return 0

    def clear_memory(self):
        for key in self.memory.keys():
            self.memory[key].clear()

    @abstractmethod
    def create_dataloader(self, batch_size):
        """
        creates dataset and dataloader for training.
        """
        pass

    def state(self, vega: VegaServiceNull) -> LAMarketState:
        position = self.vega.positions_by_market(self.wallet_name, self.market_id)

        position = position[0].open_volume if position else 0
        account = self.vega.party_account(
            wallet_name=self.wallet_name,
            asset_id=self.tdai_id,
            market_id=self.market_id,
        )
        book_state = self.vega.market_depth(
            self.market_id, num_levels=self.num_levels
        )  # make num_levels as a parameter?

        market_info = vega.market_info(market_id=self.market_id)
        fee = (
            float(market_info.fees.factors.liquidity_fee)
            + float(market_info.fees.factors.maker_fee)
            + float(market_info.fees.factors.infrastructure_fee)
        )
        init_price = self.price_process[0]
        next_price = (
            self.price_process[self.step_num + 1]
            if self.price_process is not None
            and len(self.price_process) > self.step_num + 1
            else np.nan
        )
        next_price /= init_price
        bid_prices = [level.price / init_price for level in book_state.buys] + [
            0
        ] * max(0, self.num_levels - len(book_state.buys))
        ask_prices = [level.price / init_price for level in book_state.sells] + [
            0
        ] * max(0, self.num_levels - len(book_state.sells))

        return LAMarketState(
            step=self.step_num,
            position=position,
            full_balance=(account.margin + account.general) / self.initial_balance,
            market_in_auction=(
                not market_info.trading_mode
                == markets_protos.Market.TradingMode.TRADING_MODE_CONTINUOUS
            ),
            bid_prices=bid_prices,
            ask_prices=ask_prices,
            bid_volumes=[level.volume for level in book_state.buys]
            + [0] * max(0, self.num_levels - len(book_state.buys)),
            ask_volumes=[level.volume for level in book_state.sells]
            + [0] * max(0, self.num_levels - len(book_state.sells)),
            trading_fee=fee,
            next_price=next_price,
        )

    @abstractmethod
    def empty_action(self) -> AbstractAction:
        pass

    def finalise(self):
        numTries = 3
        account = None
        for i in range(0, numTries):
            account = self.vega.party_account(
                wallet_name=self.wallet_name,
                asset_id=self.tdai_id,
                market_id=self.market_id,
            )
            self.latest_state = self.state(self.vega)
            if account.margin == 0:
                break
            self.vega.forward("1s")
            self.vega.wait_for_total_catchup()

        if account.margin > 0:
            print(
                "Market should be settled but there is still balance in margin account. What's up?"
            )

        self.latest_action = self.empty_action()
        self.step_num += 1
        # final_pnl = self.latest_state.full_balance - self.initial_balance
        final_pnl = self.latest_state.full_balance - 1.0
        with open(self.logfile_pnl, "a") as f:
            f.write("{},{:.8f}\n".format(self.lerningIteration, final_pnl))

        return super().finalise()

    @abstractmethod
    def step(self, vega_state: VegaState):
        pass

    @abstractmethod
    def save(self, results_dir: str):
        pass

    @abstractmethod
    def load(self, results_dir: str):
        pass
